"""Build a VISjs view of a System."""
import os
from typing import Dict, List, Optional, Tuple, Union

from cosapp.ports.port import ExtensiblePort
from cosapp.drivers import Driver
from cosapp.systems import System


def to_visjs(system: System, filename: str, embeddable: bool = False):
    """Export a system as HTML using vis.JS library.

    Parameters
    ----------
    system : System
        System to export
    filename : str
        Filename to write to
    embeddable: bool, optional
        Is the HTML to be embedded in an existing page? Default: False
    """
    # Optional dependencies
    try:
        from jinja2 import Environment, PackageLoader
    except ImportError:
        raise ImportError("jinja2 needs to be installed to export to vis JS.")

    env = Environment(
        loader=PackageLoader("cosapp.tools", "templates"),
        trim_blocks=True,
        lstrip_blocks=True,
    )
    template = env.get_template("system_repr.html")

    # This is inspired from openmdao problem_viewer

    html_begin_tags = """<!DOCTYPE html>
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
<title>{{title}}</title>
</head>
<body>\n
"""
    html_end_tags = """
</body>
</html>
"""

    from cosapp.tools import templates

    ressource_dir = os.path.dirname(os.path.abspath(templates.__file__))
    libs_dir = os.path.join(ressource_dir, "lib")

    # Grab the ressources
    with open(os.path.join(libs_dir, "vis.min.js"), "rb") as f:
        vis_js = f.read().decode("utf-8")
    with open(os.path.join(libs_dir, "vis.min.css"), "rb") as f:
        vis_css = f.read().decode("utf-8")

    def to_visJS(system: System, focus_system: Optional[System] = None) -> Dict:
        """Convert the `System` in a dictionary with 4 keys: title, nodes, edges and groups.

        The title is a string with the Module name.
        The nodes is a list of dictionaries defining node for vis JS library.
        The edges is a list of dictionaries defining edge for vis JS library.
        The groups is a list of string defining the nodes clusters for vis JS library.

        Parameters
        ----------
        focus_system : System or None
            System on which the visualization should focus by default

        Returns
        -------
        dict
            Dictionary containing elements to draw the `System` using visJS library
        """

        if system.parent is not None:
            center = system if focus_system is None else focus_system
            return to_visJS(system.parent, center)

        def get_component2id(
            system: System,
            cmpt2id: Dict[Union[System, ExtensiblePort], int],
            node_id: int,
            edges: List[Dict],
        ) -> Tuple[Dict[Union[System, ExtensiblePort], int], int, List[Dict]]:
            """Set the mapping between the `Module` and their unique id.

            Parameters
            ----------
            system : Module
                The parent `Module` for which id must be set
            cmpt2id : Dict[Union[Module, ExtensiblePort], int]
                Mapping of `Module` or port and its id
            node_id : int
                Latest used id
            edges : List[Dict]
                List of connection edges

            Returns
            -------
            tuple (List[Dict], Dict[Union['Module', 'ExtensiblePort'], int], int)
                The first is the mapping between each involved system or port and an unique id,
                the second is the latest used id and
                the third index is the list of edges (one per connector).
            """

            def get_driver2id(driver, cmpt2id, node_id, edges):
                if driver not in cmpt2id:
                    cmpt2id[driver] = node_id
                    node_id = node_id + 1

                previous_child = None
                for child in driver.children.values():
                    cmpt2id, node_id, edges = get_driver2id(
                        child, cmpt2id, node_id, edges
                    )
                    if (
                        previous_child is not None
                    ):  # Connection between same level driver
                        edges.append(
                            {
                                "from": cmpt2id[previous_child],
                                "to": cmpt2id[child],
                                "arrows": "to",
                                "title": "",
                                "hidden": True,
                                "physics": False,
                            }
                        )
                    previous_child = child
                    # Connection with parent driver
                    edges.append(
                        {
                            "from": cmpt2id[driver],
                            "to": cmpt2id[child],
                            "title": "",
                            "hidden": True,
                            "physics": False,
                        }
                    )

                return cmpt2id, node_id, edges

            def connectors_to_visJS(
                system: System,
                component_id: Optional[Dict[Union[System, ExtensiblePort], int]] = None,
                node_id: int = 1,
            ) -> Tuple[List[Dict], Dict[Union[System, ExtensiblePort], int], int]:
                """Build the dictionary the list of edges.

                Parameters
                ----------
                system : System
                    System owning the connectors
                component_id : Dict[Union[System, Port], int], optional
                    Mapping of `System` or `Port` and its unique id; default None
                node_id : int, optional
                    Latest used id; default 1

                Returns
                -------
                tuple (List[Dict], Dict[Union['System', 'ExtensiblePort'], int], int)
                    The first index is the list of edges (one per connector),
                    the second is the mapping between each involved system or port and an unique id and
                    the third is the latest used id.
                """
                edges = list()

                for connection in system.connectors.values():
                    supplier = connection.source.owner
                    if len(supplier.children):  # Insert port as node
                        supplier = connection.source

                    if supplier not in component_id:
                        component_id[supplier] = node_id
                        node_id = node_id + 1

                    target = connection.sink.owner
                    if len(target.children):  # Insert port as node
                        target = connection.sink

                    if target not in component_id:
                        component_id[target] = node_id
                        node_id = node_id + 1

                    edge = {
                        "from": component_id[supplier],
                        "to": component_id[target],
                        "arrows": "to",
                        "title": "",
                    }
                    if set(connection.variable_mapping.values()) == set(
                        connection.variable_mapping
                    ):
                        edge["title"] = str(list(connection.variable_mapping))
                    else:
                        edge["title"] = str(connection.variable_mapping)

                    edges.append(edge)

                    # if supplier or target is top-system, set edge.length = 0
                    if (
                        connection.source.owner is connection.sink.owner.parent
                        or connection.source.owner.parent is connection.sink.owner
                    ):
                        edges[-1]["length"] = 0

                return edges, component_id, node_id

            for driver in system.drivers.values():
                cmpt2id, node_id, edges = get_driver2id(driver, cmpt2id, node_id, edges)

            for cpt in system.children.values():
                if cpt not in cmpt2id:
                    cmpt2id[cpt] = node_id
                    node_id = node_id + 1

                if len(cpt.children):
                    cmpt2id, node_id, edges = get_component2id(
                        cpt, cmpt2id, node_id, edges
                    )

            local_edges, cmpt2id, node_id = connectors_to_visJS(system, cmpt2id, node_id)
            edges.extend(local_edges)

            return cmpt2id, node_id, edges

        cmpt2id = dict()
        id = 1
        edges = list()
        cmpt2id, id_init, edges = get_component2id(system, cmpt2id, id, edges)

        nodes = list()
        # Add hidden node for the top system
        nodes.append(
            dict(
                label=f"{system.name}",
                title=f"{system.name} - {system.__class__.__name__}",
                id=id_init,
                level=0,
                hidden=True,
                physics=False,
                group=f".{system.name}",
            )
        )

        # Add edges from uppest level to its children but deactivate physics on
        # them. The goal is to have them available for the hierarchical view but
        # not playing any role in the default layout.
        for child in system.children.values():
            edges.append(
                {
                    "from": id_init,
                    "to": cmpt2id[child],
                    "hidden": True,
                    "physics": False,
                }
            )

        def get_fullname(system: System) -> str:
            name = [system.name]
            parent = system
            while parent.parent is not None:
                parent = parent.parent
                name.insert(0, parent.name)

            return ".".join(name)

        groups = list()
        for c, id in cmpt2id.items():
            if isinstance(c, ExtensiblePort):
                port_name = c.name

                ref = dict(
                    title=f"{c.owner.name}.{port_name} - {c.__class__.__name__}",
                    id=id,
                    group=f"{get_fullname(c.owner)}",
                    mass=2,
                )

                # Add edge to bind to central owner node
                edge = {
                    "from": id,
                    "to": cmpt2id[c.owner] if c.owner in cmpt2id else id_init,
                    "hidden": True,
                }
                if id_init in (edge["from"], edge["to"]):
                    edge["physics"] = False
                edges.append(edge)

            elif isinstance(c, Driver):
                # Add driver node
                ref = dict(
                    label=type(c).__name__,
                    title=f"{c.name} - {type(c).__name__}",
                    id=id,
                    level=-0.7,
                    shape="box",
                    hidden=True,
                    physics=False,
                    group=f"{get_fullname(c.owner)}",
                )

                # Add edge to Module owner if top driver
                if c.parent is None:
                    edges.append(
                        {
                            "from": cmpt2id[c.owner] if c.owner in cmpt2id else id_init,
                            "to": cmpt2id[c],
                            "dashes": True,
                            "hidden": True,
                            "physics": False,
                        }
                    )
                else:

                    def driver_depth(driver: Driver, depth: int = 0) -> int:
                        if driver.parent is None:
                            return depth
                        else:
                            return driver_depth(driver.parent, depth + 1)

                    ref["level"] += 0.2 * driver_depth(c)

            else:
                ref = dict(
                    label=f"{c.name}",
                    title=f"{c.parent.name}.{c.name} - {c.__class__.__name__}", 
                    id=id,
                )

                if len(c.children) > 0:  # Component containing component => Cluster
                    ref["hidden"] = True

                    # Add edge to force binding to central system node
                    for child in c.children.values():
                        edges.append({"from": id, "to": cmpt2id[child], "hidden": True})

                ref["group"] = f"{get_fullname(c.parent)}"

            if "level" not in ref:
                ref["level"] = 0
            ref["level"] += ref["group"].count(".") + 1  # Set hierarchical level

            nodes.append(ref)
            if "group" in ref and ref["group"] not in groups:
                groups.append(ref["group"])

        # Sort the groups by system to ensure that levels are clustered bottom-up
        groups.sort(reverse=True)

        data = {"title": system.name, "nodes": nodes, "edges": edges, "groups": groups}

        if focus_system is not None:
            data["center_id"] = cmpt2id[focus_system]

        return data

    elements = to_visJS(system)
    elements["visJS"] = vis_js
    elements["visCSS"] = vis_css

    rendered_html = template.render(elements)
    if not embeddable:
        html_begin_tags = html_begin_tags.replace("{{title}}", elements["title"])
        rendered_html = html_begin_tags + rendered_html + html_end_tags

    _, ext = os.path.splitext(filename)
    if ext != ".html":
        filename = f"{filename}{os.extsep}html"

    with open(filename, mode="w", encoding="utf-8") as f:
        f.write(rendered_html)



from typing import Union
import numpy as np
from cosapp.systems import System
from cosapp.tests.library.ports import V1dPort, V2dPort


class Strait1dLine(System):
    def setup(self, **kwargs):
        self.add_input(V1dPort, "in_")
        self.add_inward("a", np.ones(3))
        self.add_inward("b", np.zeros(3))
        self.add_output(V1dPort, "out")

    def compute(self):
        self.out.x = self.a * self.in_.x + self.b


class Strait2dLine(System):
    def setup(self, **kwargs):
        self.add_input(V2dPort, "in_")
        self.add_inward("a", np.ones(3, 3))
        self.add_inward("b", np.zeros(3, 3))
        self.add_output(V2dPort, "out")

    def compute(self):
        self.out.x = self.a * self.in_.x + self.b


class Splitter1d(System):
    def setup(self, **kwargs):
        self.add_input(V1dPort, "in_")
        self.add_inward("s", 0.1 * np.ones(3))
        self.add_output(V1dPort, "out1")
        self.add_output(V1dPort, "out2")

    def compute(self):
        self.out1.x = self.s * self.in_.x
        self.out2.x = (1. - self.s) * self.in_.x


class Merger1d(System):
    def setup(self, **kwargs):
        self.add_input(V1dPort, "in1")
        self.add_input(V1dPort, "in2")
        self.add_output(V1dPort, "out")

    def compute(self):
        self.out.x = self.in1.x + self.in2.x


class AllTypesSystem(System):
    def setup(self, **kwargs):
        dim = 3
        self.add_property("dimension", dim)
        self.add_input(V1dPort, "in_")
        self.add_inward("a", np.ones(dim), unit="kg")
        self.add_inward("b", np.zeros(dim), unit="N")
        self.add_inward("c", 23, unit="m")
        self.add_inward("e", "sammy")
        self.add_outward("d", list())
        self.add_output(V1dPort, "out")

    def compute(self):
        self.out.x = self.a * self.in_.x + self.b
        self.d = [self.c, self.e]


class BooleanSystem(System):
    def setup(self, **kwargs):
        self.add_input(V1dPort, "in_")
        self.add_inward("a", True)




import logging
import numpy

from cosapp.drivers.runonce import RunOnce
from cosapp.core.numerics.basics import MathematicalProblem

logger = logging.getLogger(__name__)


class IterativeCase(RunOnce):
    """Abstract interface to children cases for a :py:class:`~cosapp.drivers.abstractsolver.AbstractSolver`.

    Parameters
    ----------
    name: str, optional
        Name of the `Module`
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
    **kwargs : Dict[str, Any]
        Optional keywords arguments
    """

    __slots__ = ('design')

    def __init__(
        self, name: str, owner: "Optional[cosapp.systems.System]" = None, **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)

        self.design = MathematicalProblem(self.name, self.owner)  # type: MathematicalProblem
            # desc="Additional mathematical problem to solve for the case.",

    @RunOnce.owner.setter
    def owner(self, value: "Optional[cosapp.systems.System]") -> None:
        # Trick to call super setter (see: https://bugs.python.org/issue14965)
        if self.owner is not value:
            if self.owner is not None:
                logger.warning(
                    f"System owner of Driver {self.name!r} has changed. Optimization equations have been cleared."
                )
            self.design = MathematicalProblem(self.design.name, value)
        super(IterativeCase, IterativeCase).owner.__set__(self, value)

    def set_iteratives(self, x: numpy.ndarray) -> int:
        """Set iteratives from the vector x.

        Parameters
        ----------
        x : numpy.ndarray
            The non-consumed new iteratives vector

        Returns
        -------
        int
            The number of value consumed.
        """
        counter = 0
        for name, unknown in self.get_problem().unknowns.items():
            if unknown.mask is None:
                unknown.set_default_value(x[counter])
                counter += 1
            else:
                n = numpy.count_nonzero(unknown.mask)
                unknown.set_default_value(x[counter : counter + n])
                counter += n
            # Set all design variables at once
            if name in self.design.unknowns:
                # Set the variable to the new x
                if not numpy.array_equal(unknown.value, unknown.default_value):
                    unknown.set_to_default()

        return counter

    def _postcompute(self) -> None:
        """Actions to carry out after the :py:meth:`~cosapp.drivers.runonce.RunOnce.compute` method call.

        This gathers the residues for this point and undo the variable status changes
        """
        # Request the residues of the current case to be updated
        for residue in self.get_problem().residues.values():
            residue.update()

        super()._postcompute()

    # TODO Fred do we need a get_problem here or better integration between IterativeCase -> RunOptim & RunSingleCase



from numbers import Number

from cosapp.drivers.time.interfaces import ExplicitTimeDriver


class EulerExplicit(ExplicitTimeDriver):
    def __init__(self, name="", owner: "Optional[cosapp.systems.System]" = None, **options):
        """Initialization of the driver

        Parameters
        ----------
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belongs; default None
        **options : Dict[str, Any]
            Optional keywords arguments; may contain time step and interval, with keys `dt` and `time_interval`
        """
        name = name or self.algo_info
        super().__init__(name, owner, **options)

    @property
    def algo_info(self) -> str:
        """str: Short description of driver's algorithm"""
        return "Explicit Euler time driver"

    def _update_transients(self, dt: Number) -> None:
        """
        Time integration of transient variables over time step `dt` by explicit Euler scheme.
        """
        for x in self._transients.values():
            x.value += x.d_dt * dt



import logging
from collections import OrderedDict
from numbers import Number
from typing import Any, Dict, Iterable, Optional, Sequence, Union, Tuple

import numpy

from cosapp.core.variableref import VariableReference
from cosapp.core.numerics.boundary import Unknown, TimeUnknown, TimeDerivative
from cosapp.core.numerics.residues import Residue

logger = logging.getLogger(__name__)


class SolverResults:
    """Storage class for solver solution

    Attributes
    ----------
    x : Sequence[float]
        Solution vector
    success : bool
        Whether or not the solver exited successfully.
    message : str
        Description of the cause of the termination.
    fun : ndarray
        Values of objective function.
    jac_lup : (ndarray[float], ndarray[int]), optional
        LU decomposition of Jacobian given as tuple (LU, perm); (None, None) if not available.
    jac : ndarray, optional
        Values of function Jacobian; None if not available.
    jac_errors : dict, optional
        Dictionary with singular Jacobian matrix error indexes
    jac_calls : int
        Number of calls to the Jacobian matrix calculation
    fres_calls : int
        Number of calls to the residues function
    trace : List[Dict[str, Any]]
        Resolution history (if requested) with the evolution of x, residues and jacobian
    """

    def __init__(self):
        self.x = list()  # type: Sequence[float]
        self.success = False  # type: bool
        self.message = ''  # type: str
        self.fun = numpy.array([], dtype=float)  # type: numpy.ndarray
        self.jac_lup = (None, None)  # type: Optional[(numpy.ndarray, numpy.ndarray)]
        self.jac = None  # type: Optional[numpy.ndarray]
        self.jac_errors = dict()
        self.jac_calls = 0  # type: int
        self.fres_calls = 0  # type: int
        self.trace = list()  # type: List[Dict[str, Any]]


class MathematicalProblem:
    """Container object for unknowns and equations.

    Parameters
    ----------
    name : str
        Name of the mathematical problem
    context : cosapp.systems.System
        Context in which the mathematical problem will be evaluated.
    """

    def __init__(self, name: str, context: 'Optional[cosapp.systems.System]') -> None:
        # TODO add point label to associate set of equations with Single Case
        self._name = name  # type: str
        self._context = context  # type: Optional[cosapp.systems.System]
        self._unknowns = OrderedDict()  # type: Dict[str, Unknown]
        self._residues = OrderedDict()  # type: Dict[str, Residue]
        self._transients = OrderedDict()  # type: Dict[str, TimeUnknown]
        self._rates = OrderedDict()  # type: Dict[str, TimeDerivative]

    def __str__(self) -> str:
        msg = ""
        if len(self.unknowns) > 0:
            msg += "Unknowns\n"
            for name in self.unknowns:
                msg += "  " + name + "\n"
        if len(self.residues) > 0:
            msg += "Equations\n"
            for residue in self.residues.values():
                msg += "  " + str(residue) + "\n"
        return msg

    def __repr__(self) -> str:
        msg = ""
        if len(self.unknowns) > 0:
            msg += "Unknowns\n"
            for unknown in self.unknowns.values():
                msg += "  " + repr(unknown) + "\n"
        if len(self.residues) > 0:
            msg += "Equations\n"
            for residue in self.residues.values():
                msg += "  " + str(residue) + "\n"
        return msg

    @property
    def name(self) -> str:
        """str : Mathematical system name."""
        return self._name

    @property
    def context(self) -> 'Optional[cosapp.systems.System]':
        """cosapp.systems.System or None : Context in which the mathematical system is evaluated."""
        return self._context

    @context.setter
    def context(self, context: 'Optional[cosapp.systems.System]'):
        if self.context is None:
            self._context = context
        else:
            raise ValueError(f"Context is already set to {self.context.name!r}.")

    @property
    def residues(self) -> Dict[str, Residue]:
        """Dict[str, Residue] : Residue equations."""
        return self._residues

    @property
    def residues_names(self) -> Tuple[str]:
        """Tuple[str] : Names of residues, flattened to have the same size as `residues_vector`."""
        names = []
        for name, residue in self.residues.items():
            n_values = numpy.size(residue.value)
            if n_values > 1:
                names.extend([f"{name}[{i}]" for i in range(n_values)])
            else:
                names.append(name)
        return tuple(names)

    @property
    def residues_vector(self) -> numpy.ndarray:
        """numpy.ndarray : Vector of residues."""
        residues = numpy.empty(0)
        for residue in self.residues.values():
            residues = numpy.append(residues, residue.value)
        return residues

    @property
    def shape(self) -> Tuple[int, int]:
        """(int, int) : Number of unknowns and equations."""
        n_unknowns = 0
        for unknown in self.unknowns.values():
            n_unknowns += numpy.size(unknown.value)
        return n_unknowns, self.residues_vector.size

    @property
    def unknowns(self) -> Dict[str, Unknown]:
        """Dict[str, Unknown] : Unknown numerical features defined for this system."""
        return self._unknowns

    @property
    def unknowns_names(self) -> Tuple[str]:
        """Tuple[str] : Names of unknowns flatten to have the same size as `residues_vector`."""
        names = []
        for name, unknown in self.unknowns.items():
            n_values = numpy.size(unknown.value)
            if n_values > 1:
                names.extend([f"{name}[{i}]" for i in numpy.arange(n_values)[unknown.mask]])
            else:
                names.append(name)
        return tuple(names)

    def add_unknown(
            self,
            name: Union[str, Iterable[Union[dict, str, Unknown]]],
            max_abs_step: Number = numpy.inf,
            max_rel_step: Number = numpy.inf,
            lower_bound: Number = -numpy.inf,
            upper_bound: Number = numpy.inf,
            mask: Optional[numpy.ndarray] = None,
    ) -> "MathematicalProblem":
        """Add unknown variables.

        You can set variable one by one or provide a list of dictionary to set multiple variable at once. The
        dictionary key are the arguments of this method.

        Parameters
        ----------
        name : str or Iterable of dictionary or str
            Name of the variable or list of variable to add
        max_rel_step : float, optional
            Maximal relative step by which the variable can be modified by the numerical solver; default numpy.inf
        max_abs_step : float, optional
            Maximal absolute step by which the variable can be modified by the numerical solver; default numpy.inf
        lower_bound : float, optional
            Lower bound on which the solver solution is saturated; default -numpy.inf
        upper_bound : float, optional
            Upper bound on which the solver solution is saturated; default numpy.inf
        mask : numpy.ndarray or None
            Mask of unknown values in the vector variable

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """

        def add_unknown(
            context: 'cosapp.systems.System',
            name: str,
            max_abs_step: Number = numpy.inf,
            max_rel_step: Number = numpy.inf,
            lower_bound: Number = -numpy.inf,
            upper_bound: Number = numpy.inf,
            mask: Optional[numpy.ndarray] = None,
        ):
            """Set numerical limitations on variables.

            Parameters
            ----------
            context: cosapp.systems.System
                Object in which the unknown is defined
            name : str
                Name of the variable
            max_rel_step : float, optional
                Maximal relative step by which the variable can be modified by the numerical solver; default numpy.inf
            max_abs_step : float, optional
                Maximal absolute step by which the variable can be modified by the numerical solver; default numpy.inf
            lower_bound : float, optional
                Lower bound on which the solver solution is saturated; default -numpy.inf
            upper_bound : float, optional
                Upper bound on which the solver solution is saturated; default numpy.inf
            mask : numpy.ndarray or None
                Mask of unknown values in the vector variable
            """
            # TODO we have a problem here if a vector variable is defined as unknown partially multiple times
            #   Example a = [1, 2, 3] with Unknown1 = a[0] & Unknown2 = a[2]

            unknown = Unknown(context, name, max_abs_step, max_rel_step, lower_bound, upper_bound, mask)
            if unknown.name in self._unknowns:
                logger.info(
                    f"Variable {name!r} is already declared as unknown in {self.name!r}."
                )
            else:
                self._unknowns[unknown.name] = unknown

        if self.context is None:
            raise AttributeError("Owner System is required to define unknowns.")

        if isinstance(name, str):
            add_unknown(self.context, name, max_abs_step, max_rel_step, lower_bound, upper_bound, mask)
        else:
            for unknown in name:
                if isinstance(unknown, Unknown):
                    current_to_context = self.context.get_path_to_child(unknown.context)
                    new_name = f"{current_to_context}.{name}" if current_to_context else unknown.name
                    if new_name in self._unknowns:
                        logger.warning(
                            "Unknown {!r} already exists in mathematical system {!r}. "
                            "It will be overwritten.".format(new_name, self.name)
                        )
                    self._unknowns[new_name] = unknown
                elif isinstance(unknown, str):
                    add_unknown(self.context, unknown, max_abs_step, max_rel_step, lower_bound, upper_bound, mask)
                else:
                    add_unknown(self.context, **unknown)

        return self

    def add_equation(self,
        equation: Union[str, Iterable[Union[dict, str, Tuple[str, str]]]],
        name: Optional[str] = None,
        reference: Union[Number, numpy.ndarray, str] = 1,
    ) -> "MathematicalProblem":
        """Add residue equation.

        You can add residue equation one by one or provide a list of dictionary to add multiple equation at once. The
        dictionary key are the arguments of this method.

        Parameters
        ----------
        equation : str or Iterable of str of the kind 'lhs == rhs'
            Equation or list of equations to add
        name : str, optional
            Name of the equation; default None => 'lhs == rhs'
        reference : Number, numpy.ndarray or "norm", optional
            Reference value(s) used to normalize the equation; default is 1.
            If value is "norm", actual reference value is estimated from order of magnitude.

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """

        if self.context is None:
            raise AttributeError("Owner System is required to define equations.")

        def add_residue(equation, name=None, reference=1):
            """Add residue from equation."""
            residue = Residue(self.context, equation, name, reference)
            self._residues[residue.name] = residue

        if isinstance(equation, str):
            add_residue(equation, name, reference)
        else:
            for eq in equation:
                if isinstance(eq, str):
                    add_residue(eq)
                elif isinstance(eq, dict):
                    add_residue(**eq)
                else:
                    add_residue(*eq)

        return self

    @property
    def transients(self) -> Dict[str, TimeUnknown]:
        """Dict[str, TimeUnknown] : Unknown time-dependent numerical features defined for this system."""
        return self._transients

    def add_transient(self,
        name: str,
        der: Any,
        max_time_step: Union[Number, str] = numpy.inf,
        max_abs_step: Union[Number, str] = numpy.inf,
        pulled_from: Optional[VariableReference] = None,
    ) -> "MathematicalProblem":
        """Add a time-dependent unknown.

        Parameters
        ----------
        name : str
            Name of the new time-dependent (transient) variable
        der : Any
            Name of the variable or evaluable expression defining the time derivative of transient variable
        max_time_step : Number or evaluable expression (str), optional
            Maximal time step for the time integration of the transient variable; numpy.inf by default.
        max_abs_step : Number or evaluable expression compatible with transient variable, optional
            Maximum variable step admissible over one time step; numpy.inf by default.
        pulled_from : VariableReference, optional
            Original time unknown before pulling; None by default.

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """

        if self.context is None:
            raise AttributeError("Owner System is required to define time-dependent unknowns.")

        if name in self._transients:
            raise ArithmeticError(f"Variable {name!r} is already defined as a time-dependent unknown of {self.name!r}.")

        self._transients[name] = TimeUnknown(self.context, name, der, max_time_step, max_abs_step, pulled_from=pulled_from)
        return self

    @property
    def rates(self) -> Dict[str, TimeDerivative]:
        """Dict[str, TimeDerivative] : Time derivatives computed during system evolution."""
        return self._rates

    def add_rate(self, name: str, source: Any, initial_value: Any = None) -> "MathematicalProblem":
        """Add a time derivative.

        Parameters
        ----------
        name : str
            Name of the new time-dependent (transient) variable
        source : Any
            Name of the variable or evaluable expression whose time derivative will be computed

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """

        if self.context is None:
            raise AttributeError("Owner System is required to define a time derivative.")

        if name in self._rates:
            raise ArithmeticError(f"Variable {name!r} is already defined as a time-dependent unknown of {self.name!r}.")

        self._rates[name] = TimeDerivative(self.context, name, source, initial_value)
        return self

    def extend(self, other: 'MathematicalProblem', copy: bool = True) -> 'MathematicalProblem':
        """Extend the current mathematical system with the other one.

        Parameters
        ----------
        other : MathematicalProblem
            The other mathematical system to add
        copy : bool, optional
            Should the objects be copied; default True

        Returns
        -------
        MathematicalProblem
            The resulting mathematical system
        """
        current_to_context = self.context.get_path_to_child(other.context)

        if len(current_to_context) > 0:
            full_path = lambda name: f"{current_to_context}.{name}"
            residue_fullname = lambda name: full_path(name if name.endswith(')') else name.join('()'))
        else:
            full_path = residue_fullname = lambda name: name

        get = (lambda obj: obj.copy()) if copy else (lambda obj: obj)

        def connect(self_dict, other_dict, get_fullname):
            for name, elem in other_dict.items():
                fullname = get_fullname(name)
                if fullname in self_dict:
                    raise ValueError(f"{fullname!r} already exists in system {self.name!r}.")
                self_dict[fullname] = get(elem)

        connect(self._unknowns, other.unknowns, full_path)
        connect(self._residues, other.residues, residue_fullname)
        connect(self._transients, other.transients, full_path)
        connect(self._rates, other.rates, full_path)

        return self

    def clear(self) -> None:
        """Clear all mathematical elements in this problem."""
        self._unknowns.clear()
        self._residues.clear()
        self._transients.clear()
        self._rates.clear()
        # self._extrema.clear()

    def copy(self) -> 'MathematicalProblem':
        """Copy the `MathematicalSystem` object.

        Returns
        -------
        MathematicalProblem
            The duplicated mathematical problem.
        """
        new = MathematicalProblem(self.name, self.context)
        new.extend(self)
        return new

    def to_dict(self) -> Dict[str, Any]:
        """Returns a JSONable representation of the mathematical problem.
        
        Returns
        -------
        Dict[str, Any]
            JSONable representation
        """
        return {
            "unknowns": dict([(name, unknown.to_dict()) for name, unknown in self.unknowns.items()]),
            "equations": dict([(name, equation.to_dict()) for name, equation in self.residues.items()]),
            "transients": dict([(name, transient.to_dict()) for name, transient in self.transients.items()]),
            "rates": dict([(name, rate.to_dict()) for name, rate in self.rates.items()])
        }

    def validate(self) -> None:
        """Verifies that there are as much unknowns as equations defined.

        Raises
        ------
        ArithmeticError
            If the mathematical system is not closed.
        """
        n_unknowns, n_equations = self.shape
        if n_unknowns != n_equations:
            msg = "Nonlinear problem {} error: Mismatch between numbers of params [{}] and residues [{}]".format(
                self.name, n_unknowns, n_equations
            )
            logger.error(msg)
            logger.error(f"Residues: {list(self.residues)}")
            logger.error(f"Variables: {list(self.unknowns)}")
            raise ArithmeticError(msg)



import abc
import inspect
import weakref


class Observer(abc.ABC):
    """Generic interface for observers"""
    def __init__(self, subject=None):
        self._subject = None
        if subject is not None:
            self.observe(subject)

    @abc.abstractmethod
    def _update(self, *args, **kwargs) -> None:
        """Action to perform when notified by observed subject"""
        pass

    def observe(self, subject) -> None:
        """Sign in as observer of subject"""
        self.quit()
        subject.add(self)
        self._subject = weakref.ref(subject)

    def quit(self) -> None:
        """Quit observing whoever observer is currently observing"""
        if self._subject is not None:
            self._subject().remove(self)
        self._subject = None

    def observes(self, subject=None) -> bool:
        """Bool: does observer observe `subject`?
        If `subject` is None, returns True if observer observes anyone, False otherwise."""
        observing = (self._subject is not None)  # does observer observe anyone?
        if subject is None:
            return observing
        return observing and (self._subject() is subject)

    def __del__(self) -> None:
        try:
            object.__getattribute__(self, "_subject")
        except AttributeError:
            pass  # Quit should be skipped if `_subject` is already deleted
        else:
            self.quit()


class Subject:
    """
    Prototype of subject for Observer objects.
    The philosophy is that observers are responsible for signing in or out.
    Therefore, the subject should not unilaterally add or remove observers, except when it is cleared.
    """
    __obs_type = Observer

    def __new__(cls, obs_type=Observer):
        """
        Constructor of class `Subject`, accepting observers of type `Observer`.
        Derived classes can be constructed with a more specialized
        definition of allowed observer types, by specifying `obs_type`.
        """
        cls.__obs_type = obs_type if inspect.isclass(obs_type) else type(obs_type)
        return super().__new__(cls)

    def __init__(self):
        self._observers = weakref.WeakSet()
    
    @classmethod
    def observer_type(cls) -> type:
        """Returns the type of observers allowed to observe Subject"""
        return cls.__obs_type
    
    def notify(self, *args, **kwargs) -> None:
        """Notify observers that they must update"""
        for observer in self._observers:
            observer._update(*args, **kwargs)
    
    def add(self, observer):
        """
        Add an observer to the list of observers.
        Invoked by outside observers when they sign in; should not be called by self directly.
        """
        otype = self.observer_type()
        if not isinstance(observer, otype):
            cls_name = self.__class__.__name__
            raise TypeError(f"{cls_name} can only be observed by objects of type {otype.__name__}")
        self._observers.add(observer)

    def remove(self, observer) -> None:
        self._observers.remove(observer)

    def clear(self) -> None:
        """Force all observers to quit"""
        for observer in self._observers.copy():
            observer.quit()

    @property
    def n_observers(self) -> int:
        """int: number of observers of current subject"""
        return len(self._observers)

    def __del__(self) -> None:
        self.clear()



import logging

from cosapp.utils.validate import validate
from cosapp.drivers.optionaldriver import OptionalDriver

logger = logging.getLogger(__name__)


# TODO
# [ ] Quid for vector variables
class ValidityCheck(OptionalDriver):
    """
    When executed, this driver reports in the log the validity status for all variables of the
    driver `System` owner (and its children).

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
    **kwargs : Any
        Keyword arguments will be used to set driver options
    """
    
    __slots__ = ('warnings', 'errors', )

    def __init__(
        self, name: str, owner: "Optional[cosapp.systems.System]" = None, **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)
        self.warnings = dict()  # type: Dict[str, str]
        self.errors = dict()  # type: Dict[str, str]

    def compute(self) -> None:
        """Report in the log the validity status for all variables of the driver `System` owner
        (and its children).
        """
        def message(log_dict):
            return "\n" + "\n\t".join([f"{key}{msg}" for key, msg in log_dict.items()])
        self.warnings, self.errors = validate(self.owner)

        if self.warnings:
            logger.warning(message(self.warnings))

        if self.errors:
            logger.error(message(self.errors))



import logging
from io import StringIO
from typing import (Any, Callable, Dict, List, Optional, Sequence,
                    Tuple, Union)

import numpy
import pandas

from cosapp.core.numerics.basics import SolverResults
from cosapp.core.numerics.enum import NonLinearMethods
from cosapp.core.numerics.root import root
from cosapp.drivers.abstractsolver import AbstractSolver
from cosapp.drivers.driver import Driver
from cosapp.drivers.iterativecase import IterativeCase
from cosapp.utils.helpers import check_arg
from cosapp.utils.logging import LogFormat, LogLevel

logger = logging.getLogger(__name__)


class NonLinearSolver(AbstractSolver):
    """Solve mathematical problem with algebraic variables.

    Attributes
    ----------
    compute_jacobian : bool
        Should the Jacobian matrix be computed? ; default True
    jac_lup : ndarray, optional
        LU decomposition of latest Jacobian matrix (if available, None otherwise)
    jac : ndarray, optional
        Latest Jacobian matrix computed (if available, None otherwise)
    """

    __slots__ = (
        '__method', '__option_aliases', '__trace',
        'compute_jacobian', 'jac_lup', 'jac'
    )

    def __init__(self, 
        name: str, 
        owner: "Optional[cosapp.systems.System]" = None, 
        method: Union[NonLinearMethods, str] = NonLinearMethods.NR, 
        **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        method : Union[NonLinearMethods, str]
            Resolution method to use
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)

        if isinstance(method, str):
            method = NonLinearMethods(method)
        self.__method = method
        self.__option_aliases = dict()
        self.__set_method(method, **kwargs)
        self.__trace: List[Dict[str, Any]] = list()

        self.compute_jacobian = True  # type: bool
            # desc='Should the Jacobian matrix be computed?'

        self.jac_lup = (None, None)  # type: Tuple[Optional[numpy.ndarray], Optional[numpy.ndarray]]
            # desc='LU decomposition of latest Jacobian matrix (if available).'
        self.jac = None  # type: Optional[numpy.ndarray]
            # desc='Latest computed Jacobian matrix (if available).'

        from cosapp.drivers.runsinglecase import RunSingleCase
        self.add_child(RunSingleCase(self._default_driver_name))

    @property
    def method(self) -> NonLinearMethods:
        """NonLinearMethods : Selected solver algorithm."""
        return self.__method

    def add_child(self, child: 'Driver', execution_index: Optional[int] = None) -> 'Driver':
        """Add a child `Driver` to the current `Driver`.

        When adding a child `Driver`, it is possible to specified its position in the execution order.

        Child `Port`, `inwards` and `outwards` can also be pulled at the parent level by providing either
        the name of the port/inward/outward or a list of them or the name mapping of the child element
        (dictionary keys) to the parent element (dictionary values).
        If the argument is not a dictionary, the name in the parent system will be the same as in the child.

        Parameters
        ----------
        child: Module
            `Module` to add to the current `Module`
        execution_index: int, optional
            Index of the execution order list at which the `Module` should be inserted;
            default is latest.

        Notes
        -----
        The added child will have its owner set to match the one of the current driver.
        """
        if len(self.children) == 1 and self._default_driver_name in self.children.keys():
            self.pop_child(self._default_driver_name)
        self.compute_jacobian = True
        return super().add_child(child, execution_index)

    def is_standalone(self) -> bool:
        """Is this Driver able to solve a system?

        Returns
        -------
        bool
            Ability to solve a system or not.
        """
        return True

    def resolution_method(self, fresidues: Callable[[Sequence[float], Union[float, str], bool], numpy.ndarray],
                          x0: Sequence[float],
                          args: Tuple[Union[float, str]] = (),
                          options: Optional[Dict[str, Any]] = None) -> SolverResults:

        if self.method == NonLinearMethods.NR:
            self.options.update(self._get_solver_limits())

        this_options = dict([(key, options[key]) for key in self.options if key in options])

        if self.method == NonLinearMethods.NR:
            this_options['compute_jacobian'] = self.compute_jacobian
            this_options['jac_lup'] = self.jac_lup
            this_options['jac'] = self.jac

        results = root(fresidues, x0, args=args, method=self.method, options=this_options)

        if results.jac_lup[0] is not None:
            self.jac_lup = results.jac_lup
            self.jac = results.jac
            self.compute_jacobian = False

        return results

    def _print_solution(self) -> None:  # TODO better returning a string
        """Print the solution in the log."""
        if self.options['verbose']:
            # TODO move it in MathematicalProblem
            logger.info(f"Parameters [{len(self.solution)}]: ")
            for k, v in self.solution.items():
                logger.info(f"   # {k}: {v}")

            for name, residue in self.problem.residues.items():
                value = numpy.asarray(residue.value)
                if value.ndim > 0:
                    spacing = " " * len(name)
                    msg = f"Residues [{len(self.problem.residues)}]: \n   # ({name}"
                    msg += f"\n   #  {spacing}".join([f", {v:.5g}" for v in value])
                    msg += ")\n"
                else:
                    logger.info(f"   # ({name}, {value:.5g})")

            tol = numpy.max(numpy.abs(self.problem.residues_vector))
            try:
                option = self.__option_aliases['tol']  # should never fail, in principle
                target = self.options[option]
            except:
                target = 0
            logger.debug(f" # Current tolerance {tol} for target {target}")

    def _precompute(self) -> None:
        """List all iteratives variables and get the initial values."""
        super()._precompute()

        for child in self.children.values():
            if isinstance(child, IterativeCase):
                self.problem.extend(child.get_problem(), copy=False)
                self.initial_values = numpy.append(self.initial_values, child.get_init(self.force_init))
            else:
                logger.warning(
                    f"Including Driver {child.name!r} without iteratives in Driver {self.name!r} is not numerically advised."
                )

    def compute(self) -> None:
        """Run the resolution method to find free vars values that annul residues
        """

        # Reset status
        self.status = ''
        self.error_code = '0'

        try:
            self.problem.validate()
        except ArithmeticError as err:
            self.status = 'ERROR'
            self.error_code = '9'
            raise err

        if len(self.initial_values) > 0:
            self.solution = {}

            # compute first order
            results = self.resolution_method(
                self._fresidues,
                self.initial_values,
                options=self.options,
            )

            self.__trace = getattr(results, "trace", list())

            if results.success:
                self.status = ''
                self.error_code = '0'
                logger.info(f"solver : {self.name}{results.message}")
            else:
                self.status = 'ERROR'
                self.error_code = '9'

                error_msg = f'The solver failed: {results.message}'

                error_desc = getattr(results, 'jac_errors', {})
                if error_desc:
                    if len(error_desc['unknowns']) > 0:
                        n_unknowns = len(error_desc['unknowns'])
                        unknown_idx = error_desc['unknowns']
                        unknown_names = [self.problem.unknowns_names[i] for i in unknown_idx]
                        error_msg += (f' \nThe {n_unknowns} following parameter(s) have '
                            f'no influence: {unknown_names} \n{unknown_idx}')


                    if len(error_desc['residues']) > 0:
                        n_equations = len(error_desc['residues'])
                        equation_names = [self.problem.residues_names[i] for i in error_desc['residues']]
                        error_msg += (f' \nThe {n_equations} following residue(s) are '
                            f'not influenced: {equation_names}')

                if self.parent is not None:
                    raise ArithmeticError(error_msg)
                else:
                    logger.error(error_msg)

            self.solution = dict([(key, unknown.default_value) for key, unknown in self.problem.unknowns.items()])
            self._print_solution()
        else:
            self.owner.run_children_drivers()
            logger.debug('No parameters/residues to solve. Fallback to children execution.')

        if self._recorder is not None:
            for child in self.children.values():
                child.run_once()
                self._recorder.record_state(child.name, self.status, self.error_code)

    def log_debug_message(
        self,
        handler: "HandlerWithContextFilters",
        record: logging.LogRecord,
        format: LogFormat = LogFormat.RAW
    ) -> bool:
        """Callback method on the driver to log more detailed information.
        
        This method will be called by the log handler when :py:meth:`~cosapp.utils.logging.LoggerContext.log_context`
        is active if the logging level is lower or equals to VERBOSE_LEVEL. It allows
        the object to send additional log message to help debugging a simulation.

        Parameters
        ----------
        handler : HandlerWithContextFilters
            Log handler on which additional message should be published.
        record : logging.LogRecord
            Log record
        format : LogFormat
            Format of the message

        Returns
        -------
        bool
            Should the provided record be logged?
        """
        message = record.getMessage()
        activate = getattr(record, "activate", None)
        emit_record = super().log_debug_message(handler, record, format)

        if message.endswith("call_setup_run") or message.endswith("call_clean_run"):
            emit_record = False

        elif activate == True:
            self.options["history"] = True
            emit_record = False

        elif activate == False:
            self.options["history"] = False
            emit_record = False

            message = ""
            for i, info in enumerate(self.__trace):
                if i == 0:
                    unknowns_trace = info["x"]
                    residues_trace = info["residues"]
                else:
                    unknowns_trace = numpy.vstack((unknowns_trace, info["x"]))
                    residues_trace = numpy.vstack((residues_trace, info["residues"]))

                if "jac" in info:
                    message += f"Iteration {i}\n"
                    container = StringIO()
                    numpy.savetxt(container, info["jac"], delimiter=",")
                    jacobian = container.getvalue()
                    unknowns = ", ".join(self.problem.unknowns_names)
                    message += f"New Jacobian matrix:\n,{unknowns}\n"
                    for residue, line in zip(self.problem.residues_names, jacobian.splitlines()):
                        message += f"{residue}, {line}\n"

            if len(self.__trace) > 0:
                if len(self.__trace) == 1:
                    size = self.problem.shape[0]
                    unknowns_trace = unknowns_trace.reshape((1, size))
                    residues_trace = residues_trace.reshape((1, size))
                unknowns_df = pandas.DataFrame(unknowns_trace, columns=self.problem.unknowns_names)
                container = StringIO()
                unknowns_df.to_csv(container, line_terminator="\n")
                message += f"Unknowns\n{container.getvalue()}\n"
                residues_df = pandas.DataFrame(residues_trace, columns=self.problem.residues_names)
                container = StringIO()
                residues_df.to_csv(container, line_terminator="\n")
                message += f"Residues\n{container.getvalue()}\n"

            handler.log(
                LogLevel.FULL_DEBUG,
                message,
                name=logger.name,
            )

        return emit_record

    def __set_method(self, method, **options):
        check_arg(method, 'method', NonLinearMethods)

        self.__method = method
        self.__option_aliases = dict()

        if self.__method == NonLinearMethods.NR:
            self.options.declare('tol', 1.0e-5, dtype=float, allow_none=True,
                desc='Absolute tolerance (in max-norm) for the residual.')
            self.options.declare('max_iter', 500, dtype=int,
                desc='The maximum number of iterations.')
            self.options.declare('eps', 1.0e-4, dtype=float, allow_none=True,
                desc='A suitable step length for the forward-difference approximation of the Jacobian (for fprime=None).'
                    ' If eps is smaller than machine precision u, it is assumed that the relative errors in the'
                    ' functions are of the order of u.')
            self.options.declare('factor', 1.0, dtype=float, allow_none=True, lower=1e-3, upper=1.0,
                desc='A parameter determining the initial step bound factor * norm(diag * x). Should be in the interval [0.1, 1].')
            self.options.declare('partial_jac', True, dtype=bool, allow_none=False,
                desc='Defines if partial Jacobian updates can be computed before a complete Jacobian matrix update.')
            self.options.declare('partial_jac_tries', 10, dtype=int, allow_none=False, lower=1, upper=10,
                desc='Defines how many partial Jacobian updates can be tried before a complete Jacobian matrix update.')
            self.options.declare('jac_update_tol', 0.01, dtype=float, allow_none=False, lower=0, upper=1,
                desc='Tolerance level for partial Jacobian matrix update, based on nonlinearity estimation.')
            self.options.declare('recorder', None, allow_none=True,
                desc='A recorder to store solver intermediate results.')
            self.options.declare('lower_bound', None, dtype=numpy.ndarray, allow_none=True,
                desc='Min values for parameters iterated by solver.')
            self.options.declare('upper_bound', None, dtype=numpy.ndarray, allow_none=True,
                desc='Max values for parameters iterated by solver.')
            self.options.declare('abs_step', None, dtype=numpy.ndarray, allow_none=True,
                desc='Max absolute step for parameters iterated by solver.')
            self.options.declare('rel_step', None, dtype=numpy.ndarray, allow_none=True,
                desc='Max relative step for parameters iterated by solver.')
            self.options.declare('history', False, dtype=bool, allow_none=False,
                desc='Request saving the resolution trace.')

        elif self.__method == NonLinearMethods.POWELL:
            self.__option_aliases = {
                'tol': 'xtol',
                'max_eval': 'maxfev',
            }

            self.options.declare('xtol', 1.0e-7, dtype=float,
                desc="The calculation will terminate if the relative error between two consecutive iterations is at most tol.")
            self.options.declare('maxfev', 0, dtype=int,
                desc='The maximum number of calls to the function. If zero, assumes 100 * (N + 1),'
                    ' where N is the number of elements in x0.')
            self.options.declare('eps', None, dtype=float, allow_none=True,
                desc='A suitable step length for the forward-difference approximation of the Jacobian (for fprime=None).'
                    ' If eps is less than machine precision u, it is assumed that the relative errors in the'
                    ' functions are of the order of u.')
            self.options.declare('factor', 0.1, dtype=float, lower=0.1, upper=100.,
                desc='A parameter determining the initial step bound factor * norm(diag * x). Should be in the interval [0.1, 100].')

        elif self.__method == NonLinearMethods.BROYDEN_GOOD:
            self.__option_aliases = {
                'tol': 'fatol',
                'num_iter': 'nit',
                'max_iter': 'maxiter',
                'min_rel_step': 'xtol',
                'min_abs_step': 'xatol',
            }

            self.options.declare('nit', 100, dtype=int,
                desc='Number of iterations to perform. If omitted (default), iterate unit tolerance is met.')
            self.options.declare('maxiter', 200, dtype=int,
                desc='Maximum number of iterations to make. If more are needed to meet convergence, NoConvergence is raised.')
            self.options.declare('disp', False, dtype=bool,
                desc='Print status to stdout on every iteration.')
            self.options.declare('fatol', 6e-6, dtype=float,
                desc='Absolute tolerance (in max-norm) for the residual. If omitted, default is 6e-6.')
            self.options.declare('line_search', 'armijo', dtype=str, allow_none=True,
                desc='Which type of a line search to use to determine the step '
                    'size in the direction given by the Jacobian approximation. Defaults to ‘armijo’.')
            self.options.declare('jac_options', {'reduction_method': 'svd'},
                dtype=dict, allow_none=True,
                desc='Options for the respective Jacobian approximation. restart, simple or svd')

        self._filter_options(options, self.__option_aliases)



import logging
from collections import OrderedDict
from typing import Any, Iterable, List, Optional, Set, Union

import numpy

from cosapp.core.numerics import sobol_seq
from cosapp.ports.port import ExtensiblePort, PortType
from cosapp.drivers.abstractsetofcases import AbstractSetOfCases
from cosapp.drivers.abstractsolver import AbstractSolver
from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


# TODO linearization does not support multipoint cases
# TODO Does not work for vector variables (at least partially connected one)
# TODO Does it work if a subsystem mutates to higher/lower fidelity
# TODO We don't forbid using an unknown
class MonteCarlo(AbstractSetOfCases):
    """
    This driver execute a MonteCarlo simulation on its system.
    """

    __slots__ = (
        'draws', 'linear', 'random_variables', 'responses', 'solver', 'reference_case_solution',
        'X0', 'Y0', 'A', 'perturbations'
    )

    def __init__(self,
        name: str,
        owner: "Optional[cosapp.systems.System]" = None,
        **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)
        self.draws = 200  # type: int
            # desc="Number of cases performed for Montecarlo calculations."
        self.linear = False  # type: bool
            # desc="True for linearisation of system before Montecarlo calculation. Default False."

        self.random_variables = OrderedDict()  # type: Dict[str, Tuple[VariableReference, Optional[Connector], Distribution]]
            # desc="Random variables in the system."
        self.responses = list()  # type: List[str]
            # We need a list as set is not ordered
            # desc="Variable names to study through Monte Carlo calculations."
        self.solver = None  # type: Optional[AbstractSolver]
            # desc="Solver acting. Used for re-init of case."
        self.reference_case_solution = dict()  # type: Dict[str, float]

        self.X0 = None  # type: Optional[numpy.ndarray]
            # desc="Vector of imposed disturbed values"
        self.Y0 = None  # type: Optional[numpy.ndarray]
            # desc="Vector of output evaluated disturbed values."            
        self.A = None  # type: Optional[numpy.ndarray]
            # desc="Matrice of influence of imposed disturbed values on results."
        self.perturbations = None  # type: Optional[numpy.ndarray]
            # desc="Array of perturbations applied on the system."

    def add_random_variable(self, names: Union[str, Iterable[str]]) -> None:
        """Add variable to be perturbated.

        The perturbation distribution is defined by the variable distribution details.

        ..
            from cosapp.core.numerics.distribution import Normal

            port.get_details('my_variable').distribution = Normal(worst=0.0, best=5.0)

        Parameters
        ----------
        names : Union[str, Iterable[str]]
            List of variables to be perturbated
        """
        # TODO it should be possible to set the distribution directly

        def add_unique_input_var(name: str):
            self.check_owner_attr(name)
            ref = self.owner.name2variable[name]
            port = ref.mapping

            if not isinstance(port, ExtensiblePort):
                raise TypeError(f"{name!r} is not a variable.")

            if port.direction != PortType.IN:
                raise TypeError(
                    f"{ref.key!r} is not an input variable of {port.contextual_name!r}"
                )

            distribution = port.get_details(ref.key).distribution
            if distribution is None:
                raise ValueError(
                    f"No distribution specified for '{port.contextual_name}.{ref.key}'"
                )

            # Test if the variable is connected
            connection = None
            if port.owner.parent is not None:
                connectors = port.owner.parent.connectors.values()
                for connector in filter(lambda c: c.sink is port, connectors):
                    if ref.key in connector.variable_mapping:
                        connection = connector
                        break

            self.random_variables[name] = (ref, connection, distribution)

        check_arg(names, 'names', (str, set, list))

        if isinstance(names, str):
            add_unique_input_var(names)
        else:
            for name in names:
                check_arg(name, name + " in 'names'", str)
                add_unique_input_var(name)

    def add_response(self, name: Union[str, Iterable[str]]) -> None:
        """Add a variable for which the statistical response will be calculated.

        Parameters
        ----------
        name : Union[str, Iterable[str]]
            List of variable names to add
        """
        def add_unique_response_var(name: str):
            self.check_owner_attr(name)
            if name not in self.responses:
                self.responses.append(name)

        check_arg(name, 'name', (str, set, list))

        if isinstance(name, str):
            add_unique_response_var(name)
        else:
            for n in name:
                check_arg(n, n + " in 'name'", str)
                add_unique_response_var(n)

    def _build_cases(self) -> None:
        """Build the list of cases to run during execution
        """
        self.cases = sobol_seq.i4_sobol_generate(len(self.random_variables), self.draws)

    def _precompute(self):
        """Save reference and build cases."""
        n_input = len(self.random_variables)
        n_output = len(self.responses)

        self.run_children()

        self.solver = None
        for child in self.children.values():
            if isinstance(child, AbstractSolver):
                self.solver = child

        if self.solver:
            self.reference_case_solution = child.save_solution()

        self._build_cases()

        if self.linear:  # precompute linear system
            if len(self.responses) == 0:
                raise ValueError("You need to define response variables to use MonteCarlo linear mode.")
            self.X0 = numpy.zeros(n_input)
            self.Y0 = numpy.zeros(n_output)
            self.A = numpy.zeros((n_output, n_input))

            # reference for influence matrix computation through center differentiation scheme
            for i, name in enumerate(self.random_variables):
                self.X0[i] = self.owner[name]
            for j, name in enumerate(self.responses):
                self.Y0[j] = self.owner[name]

            variation = 0.5 * (numpy.max(self.cases, axis=0) - numpy.min(self.cases, axis=0))
            for i, input_name in enumerate(self.random_variables):
                self.owner[input_name] = self.X0[i] + variation[i]
                self.run_children()

                for j, response_name in enumerate(self.responses):
                    self.A[j, i] = 0.5 * (self.owner[response_name] - self.Y0[j]) / variation[i]

                self.owner[input_name] = self.X0[i] - variation[i]
                self.run_children()

                for j, response_name in enumerate(self.responses):
                    self.A[j, i] -= 0.5 * (self.owner[response_name] - self.Y0[j]) / variation[i]

                # Restore system value
                self.owner[input_name] = self.X0[i]

            for j, name in enumerate(self.responses):
                self.Y0[j] = self.owner[name]

    def _precase(self, case_idx, case):
        """Hook to be called before running each case.
        
        Parameters
        ----------
        case_idx : int
            Index of the case
        case : Any
            Parameters for this case
        """
        super()._precase(case_idx, case)

        # Set perturbation
        self.perturbations = numpy.zeros(len(self.random_variables))
        for i, params in enumerate(self.random_variables.values()):
            ref, connector, distribution = params
            self.perturbations[i] = distribution.draw(case[i])
            if connector is None:
                ref.value += self.perturbations[i]
            else:
                connector.set_perturbation(ref.key, self.perturbations[i])

        if len(self.reference_case_solution) > 0:
            self.solver.load_solution(self.reference_case_solution)

    def compute(self) -> None:
        """Contains the customized `Module` calculation, to execute after children.
        """
        for case_idx, case in enumerate(self.cases):
            if len(case) > 0:
                self._precase(case_idx, case)
                if self.linear:
                    self.__run_linear()
                else:
                    self.run_children()
                self._postcase(case_idx, case)

    def __run_linear(self) -> None:
        """Approximate MonteCarlo simulation using partial derivatives matrix."""
        # TODO this is not great as we set variables in the system breaking its consistency.
        if len(self.responses) > 0:
            X = numpy.zeros(len(self.random_variables))
            for i, name in enumerate(self.random_variables):
                self.X0[i] = self.owner[name]

            Y = self.Y0 + numpy.matmul(self.A, X - self.X0)

            for j, name in enumerate(self.responses):
                self.owner[name] = Y[j]

    def _postcase(self, case_idx: int, case: Any):
        """Hook to be called before running each case.
        
        Parameters
        ----------
        case_idx : int
            Index of the case
        case : Any
            Parameters for this case
        """
        # Store the results
        super()._postcase(case_idx, case)

        # Remove the perturbation
        for i, params in enumerate(self.random_variables.values()):
            ref, connector = params[:2]
            if connector is None:
                ref.value -= self.perturbations[i]
            else:
                connector.set_perturbation(ref.key, -self.perturbations[i])



import pytest
from unittest.mock import MagicMock, patch

import logging
import re
import numpy as np

import cosapp.recorders as recorders
from cosapp.drivers.time import interfaces
from cosapp.drivers.time.interfaces import ExplicitTimeDriver
from cosapp.systems import System
from cosapp.utils.logging import LogFormat, LogLevel
from cosapp.utils.testing import rel_error, not_raised


@pytest.fixture(autouse=True)
def PatchExplicitTimeDriver():
    """Patch ExplicitTimeDriver to make it instanciable for tests"""
    patcher = patch.multiple(
        ExplicitTimeDriver,
        __abstractmethods__ = set(),
    )
    patcher.start()
    yield
    patcher.stop()


def test_ExplicitTimeDriver_init_default():
    driver = ExplicitTimeDriver()
    assert driver.name == "Explicit time driver"
    assert driver.owner is None
    assert driver.dt is None
    assert driver.time_interval is None


@pytest.mark.parametrize("settings, expected", [
    (dict(), dict()),
    (dict(name="John Doe"), dict(name="John Doe")),
    (dict(time_interval=(0, 1)), dict(time_interval=(0, 1))),
    (dict(dt=0.1, time_interval=(0, 1)), dict(dt=0.1, time_interval=(0, 1))),
    (dict(dt=0.1), dict(dt=0.1)),
    (dict(dt=1), dict(dt=1)),
    (dict(dt=0), dict(error=ValueError)),
    (dict(dt=-0.1), dict(error=ValueError)),
    (dict(dt="0.1"), dict(error=TypeError)),
    (dict(time_interval=(2, 1)), dict(error=ValueError)),
    (dict(time_interval=(1, 2, 3)), dict(error=ValueError)),
    (dict(time_interval=2.5), dict(error=TypeError)),
])
def test_ExplicitTimeDriver_init_args(settings, expected):
    error = expected.get('error', None)
    if error is None:
        driver = ExplicitTimeDriver(**settings)
        assert driver.dt == expected.get('dt', None)
        assert driver.time_interval == expected.get('time_interval', None)
        assert driver.name == expected.get('name', 'Explicit time driver')
    else:
        with pytest.raises(error):
            ExplicitTimeDriver(**settings)


@pytest.mark.parametrize("value, expected", [
    (10, dict(value=10)),
    (1e-2, dict(value=1e-2)),
    (1e-12, dict(value=1e-12)),
    (0, dict(error=ValueError, match="dt.*invalid value")),
    (-0.5, dict(error=ValueError, match="dt.*invalid value")),
    ("0.5", dict(error=TypeError, match="got str")),
])
def test_ExplicitTimeDriver_dt(value, expected):
    driver = ExplicitTimeDriver()
    assert driver.dt is None

    error = expected.get('error', None)
    if error is None:
        driver.dt = value
        assert driver.dt == expected['value']
    else:
        pattern = expected.get('match', None)
        with pytest.raises(error, match=pattern):
            driver.dt = value


@pytest.mark.parametrize("value, expected", [
    ((0, 1), dict(value=(0, 1))),
    ([0, 1], dict(value=(0, 1))),
    ((0.2, 0.2), dict(value=(0.2, 0.2))),
    ((1, 1.5), dict(value=(1, 1.5))),
    ((-0.5, 0), dict(error=ValueError, match="start time")),
    (('0', 0.2), dict(error=TypeError, match="start time")),
    ((0.2, 0.1), dict(error=ValueError, match="end time")),
    ((0, np.inf), dict(error=ValueError, match="end time")),
    ((0, '0.5'), dict(error=TypeError, match="end time")),
    ((1, 2, 3), dict(error=ValueError, match="time_interval.*invalid")),
    (2.5, dict(error=TypeError, match="got float")),
    ("lifetime", dict(error=TypeError, match="got str")),
])
def test_ExplicitTimeDriver_time_interval(value, expected):
    driver = ExplicitTimeDriver()
    assert driver.time_interval is None

    error = expected.get('error', None)
    if error is None:
        driver.time_interval = value
        assert driver.time_interval == expected['value']
    else:
        pattern = expected.get('match', None)
        with pytest.raises(error, match=pattern):
            driver.time_interval = value


@pytest.mark.parametrize("dt, expected", [(1e-2, 1e-2), (0.1, 0.02), (None, 0.02)])
def test_ExplicitTimeDriver_limit_dt(ode_case_1, dt, expected):
    ode, driver = ode_case_1(ExplicitTimeDriver, time_interval=[0, 0.1], record_dt=True)
    if dt is not None:
        driver.dt = dt
        assert driver.dt == dt
    else:
        assert driver.dt is None
    ode.tau = 0.1
    ode.run_drivers()
    comment = "Time step should be limited by transient max_time_step"
    assert driver.recorded_dt.min() == pytest.approx(expected, rel=1e-12), comment
    assert driver.recorded_dt.max() == pytest.approx(expected, rel=1e-12), comment


@pytest.mark.parametrize("scenario, ok", [
    # constant value of f.max_time_step
    (dict(values={'sub2.y': 10}), True),
    (dict(values={'sub2.y': 0}), False),
    (dict(values={'sub2.y': -0.1}), False),
    # non-const max_time_step
    (dict(values={'sub2.y': '1 if t < 0.2 else -1'}, time_interval=(0, 0.3)), False),
    (dict(values={'sub2.y': '1 if t < 0.5 else -1'}, time_interval=(0, 0.3)), True),
])
def test_ExplicitTimeDriver_dt_RuntimeError(scenario, ok):
    """
    Check that time driver raises RuntimeError when the maximum time step of
    one of the transients is not strictly positive at the beginning of time loop.
    """
    class SubSystem(System):
        def setup(self):
            self.add_inward('x', 1.0)
            self.add_inward('y', 1.0)

    class DynamicSystem(System):
        def setup(self):
            self.add_child(SubSystem('sub1'))
            self.add_child(SubSystem('sub2'))

            self.add_transient('f', der='sub1.x * sub1.y', max_time_step='0.5 * sub2.y')

    # set driver settings
    settings = dict(dt=0.1, time_interval=[0, 0.2])  # default
    for setting in settings:
        try:
            settings[setting] = scenario.pop(setting)
        except KeyError:
            continue

    s = DynamicSystem('s')
    driver = s.add_driver(ExplicitTimeDriver(**settings))
    driver.set_scenario(**scenario)

    if ok:
        try:
            s.run_drivers()
        except RuntimeError:
            pytest.fail("Should not raise RuntimeError")
    else:
        with pytest.raises(RuntimeError, match="non-positive value"):
            s.run_drivers()


def test_ExplicitTimeDriver_set_time_before_System_setup():

    start_t = 24.
    end_t = 25.

    class DummySystem(System):
        def setup_run(self):
            nonlocal start_t
            assert self.time == start_t

    driver = ExplicitTimeDriver()
    system = DummySystem('dummy')

    system.add_driver(driver)

    driver.time_interval = (start_t, end_t)
    driver.dt = end_t - start_t
    # Check that the start time is set as the clock time
    system.run_drivers()

    start_t = 5.0
    # Check that the new start time is set as the clock time
    driver.time_interval = (start_t, start_t + driver.dt)
    system.run_drivers()


def test_ExplicitTimeDriver_is_standalone():
    driver = ExplicitTimeDriver()
    assert driver.is_standalone()


def test_ExplicitTimeDriver_ode_run_driver(ode_case_1):
    ode, driver = ode_case_1(ExplicitTimeDriver, record_dt=True)
    assert driver.owner is ode
    assert driver.dt is None
    assert driver.time_interval is None
    with pytest.raises(ValueError, match="Time interval.*not specified"):
        ode.run_drivers()
    driver.time_interval = (0, 1)
    assert driver.dt is None
    ode.tau = 0.1
    ode.run_drivers()
    # Check that time step was set by max_time_step of transient 'ode.f'
    assert driver.recorded_dt.min() == pytest.approx(0.02, rel=1e-12)
    assert driver.recorded_dt.max() == pytest.approx(0.02, rel=1e-12)

    ode, driver = ode_case_1(ExplicitTimeDriver, dt=1e-2)
    assert driver.owner is ode
    assert driver.dt == 1e-2
    assert driver.time_interval is None
    with pytest.raises(ValueError, match="Time interval.*not specified"):
        ode.run_drivers()

    ode, driver = ode_case_1(ExplicitTimeDriver, dt=1e-2, time_interval=[0, 1])
    assert driver.owner is ode
    assert driver.dt == 1e-2
    assert driver.time_interval == (0, 1)


def test_ExplicitTimeDriver_dt_None():
    """Expected behaviour:
    ValueError exception raised when driver.dt is None, and `max_time_step` is not given by system transients"""
    class PlainSystem(System):
        def setup(self):
            self.add_inward('a', 1.0)
            self.add_transient('A', der='a')  # max_time_step unspecified

    system = PlainSystem('plain')
    driver = system.add_driver(ExplicitTimeDriver(time_interval=(0, 1)))
    assert driver.dt is None
    assert driver.time_interval == (0, 1)
    with pytest.raises(ValueError, match="Time step.*not specified.*and could not be determined from transient variables"):
        system.run_drivers()


def test_ExplicitTimeDriver_ode_add_scenario(ode_case_1):
    ode, driver = ode_case_1(ExplicitTimeDriver)
    driver.set_scenario()
    assert driver.scenario.name == "scenario"
    assert driver.scenario.context is ode
    driver.set_scenario("test")
    assert driver.scenario.name == "test"
    assert driver.scenario.context is ode
    driver.set_scenario("final")
    assert driver.scenario.name == "final"
    assert driver.scenario.context is ode

@pytest.mark.parametrize("driver_settings, period, expected", [
    (dict(), None, dict(period=None)),
    (dict(), 0.1, dict(period=0.1)),
    (dict(dt=1e-2), None, dict(period=None)),
    (dict(time_interval=[0, 2]), 0.1, dict(period=0.1)),
    (dict(dt=1e-1, time_interval=[0, 2]), 0.01, dict(period=0.01)),
    (dict(dt=1e-2, time_interval=[0, 2]), 0.1, dict(period=0.1)),
    (dict(dt=1e-2, time_interval=[0, 2]), 1, dict(period=1)),
    (dict(dt=1e-2, time_interval=[0, 2]), 100, dict(period=2)),
    (dict(dt=1e-2, time_interval=[0, 2]), 0, dict(error=ValueError, match="period.* invalid value")),
    (dict(dt=1e-2, time_interval=[0, 2]), -1, dict(error=ValueError, match="period.* invalid value")),
    (dict(dt=1e-2, time_interval=[0, 2]), '0.1', dict(error=TypeError)),
])
def test_ExplicitTimeDriver_add_recorder(two_tank_case, driver_settings, period, expected):
    """Test `add_recorder` method before driver execution"""
    system, driver = two_tank_case(ExplicitTimeDriver, **driver_settings)

    rec_options = dict(includes='tank?.height')
    error = expected.get('error', None)

    if error is None:
        driver.add_recorder(recorders.DataFrameRecorder(**rec_options), period)
        assert driver.recording_period == expected['period']
    else:
        pattern = expected.get('match', None)
        with pytest.raises(error, match=pattern):
            driver.add_recorder(recorders.DataFrameRecorder(**rec_options), period)


@pytest.mark.parametrize("driver_settings, period, expected", [
    (dict(), None, dict(error=ValueError, match="Time interval was not specified")),
    (dict(), 0.1, dict(error=ValueError, match="Time interval was not specified")),
    (dict(dt=1e-2), None, dict(error=ValueError, match="Time interval was not specified")),
    (dict(time_interval=[0, 2]), 0.1, dict(period=0.1)),
    (dict(time_interval=[0, 2]), None, dict(period=None, warnings=["all time steps will be recorded"])),
    (dict(dt=1e-1, time_interval=[0, 2]), 0.01, dict(period=0.01)),
    (dict(dt=1e-2, time_interval=[0, 2]), 0.1, dict(period=0.1)),
    (dict(dt=1e-2, time_interval=[0, 2]), 1, dict(period=1)),
    (dict(dt=1e-2, time_interval=[0, 2]), 100, dict(period=2)),
])
def test_ExplicitTimeDriver_recorder(ode_case_1, caplog, driver_settings, period, expected):
    """Test recorder behaviour during driver execution, assuming `add_recorder` is correct"""
    system, driver = ode_case_1(ExplicitTimeDriver, **driver_settings)

    driver.add_recorder(recorders.DataFrameRecorder(includes='*'), period)
    driver.set_scenario(init={'f': 0}, values={'tau': 0.5})

    error = expected.get('error', None)

    if error is None:
        # run simulation and capture potential warnings
        caplog.clear()
        with caplog.at_level(logging.WARNING, logger=interfaces.__name__):
            system.run_drivers()

        assert driver.recording_period == expected['period']

        warnings = expected.get('warnings', [])
        assert len(caplog.records) == len(warnings)
        for warning, record in zip(warnings, caplog.records):
            assert warning in record.message
    else:
        pattern = expected.get('match', None)
        with pytest.raises(error, match=pattern):
            system.run_drivers()


@pytest.mark.parametrize("dt, tol", [
    (1e-1, 1e-1),
    (1e-2, 1e-2),
])
def test_ExplicitTimeDriver_rate_singleTimeStep(rate_case_1, dt, tol):
    settings = dict(time_interval=(0, dt), dt=dt)
    system, driver = rate_case_1(ExplicitTimeDriver, **settings)
    assert driver.dt == driver.time_interval[1]

    driver.set_scenario(
        values = {'k': 1.5, 'U': 'exp(k * t)'}
    )
    system.run_drivers()

    assert system.k == 1.5
    solution = lambda t: system.k * np.exp(system.k * t)
    assert system.dU_dt == pytest.approx(solution(dt), rel=tol)


@pytest.mark.parametrize("dt, tol", [
    (1e-1, 1e-1),
    (1e-2, 1e-2),
])
def test_ExplicitTimeDriver_rate(rate_case_1, dt, tol):
    settings = dict(time_interval=(0, 1), dt=dt)
    system, driver = rate_case_1(ExplicitTimeDriver, **settings)

    driver.set_scenario(
        values = {'k': 1.5, 'U': 'exp(k * t)'}
    )

    recorder = driver.add_recorder(recorders.DataFrameRecorder(includes=['dU_dt']), period=0.1)

    system.run_drivers()
    solution = lambda t: np.exp(system.k * t) * system.k

    data = recorder.export_data()
    time = np.asarray(data['Reference'], dtype=float)
    solution = lambda t: system.k * np.exp(system.k * t)
    error = 0
    for t, dU_dt in zip(time, data['dU_dt']):
        exact = solution(t)
        error = max(error, rel_error(dU_dt, exact))
    assert error < tol


def test_ExplicitTimeDriver_rate_no_initial_value():
    """
    Same as `test_ExplicitTimeDriver_rate`, except rate `dU_dt`
    is not given any initial value.
    This test checks that `dU_dt` is correctly set by an explicit
    initial condition in `set_scenario`.
    Note: fixes bug reported in issue #193, "Bug in rate initialization".
    """
    class RateSystem(System):
        def setup(self):
            self.add_inward("k", 1.0)
            self.add_inward("U", 0.0)
            self.add_rate("dU_dt", source="U")

    system = RateSystem("system")
    driver = system.add_driver(ExplicitTimeDriver(time_interval=(0, 0.1), dt=0.1))

    driver.set_scenario(
        init = {'dU_dt': 'k'},
        values = {
            'U': 'exp(k * t)',
            'k': -3.6,  # should initialize `dU_dt`
        }
    )

    recorder = driver.add_recorder(recorders.DataFrameRecorder(includes=['dU_dt']))

    system.run_drivers()

    data = recorder.export_data()
    result = np.asarray(data['dU_dt'], dtype=float)
    assert system.k == -3.6
    assert result[0] == system.k
    assert result[1] == pytest.approx((np.exp(system.k * driver.dt) - 1) / driver.dt, rel=1e-14)


@pytest.mark.parametrize("options, recorder_period, expected", [
    (dict(), 0.1, dict(period=0.1)),
    (dict(dt=1e-2), None, dict(period=None)),
    (dict(dt=1e-2, time_interval=[0, 2]), 0.1, dict(period=0.1)),
    (dict(dt=1e-1, time_interval=[0, 2]), 1, dict(period=1)),
    (dict(dt=1e-1, time_interval=[0, 2]), 100, dict(period=2)),
])
def test_ExplicitTimeDriver_recorded_times(two_tank_case, options, recorder_period, expected):
    driver_settings = dict(dt=1e-1, time_interval=[0, 1])
    driver_settings.update(options)
    system, solver = two_tank_case(ExplicitTimeDriver, **driver_settings)

    recorder = solver.add_recorder(
        recorders.DataFrameRecorder(includes='tank1.height'), recorder_period)

    system.run_drivers()
    period = solver.recording_period
    assert period == expected['period']
    t0, tn = solver.time_interval
    data = recorder.export_data()
    time = np.asarray(data['Reference'], dtype=float)
    if period is not None:
        assert time == pytest.approx(np.arange(t0, tn + period / 2, period), abs=1e-12)


@pytest.mark.parametrize("format", LogFormat)
@pytest.mark.parametrize("msg, kwargs, to_log, emitted", [
    ("zombie call_setup_run", dict(), False, None),
    ("useless start call_clean_run", dict(activate=True), False, None),
    (
        f"{System.CONTEXT_EXIT_MESSAGE} call_clean_run",
        dict(activate=False),
        False,
        dict(levelno=LogLevel.DEBUG, pattern=r"Compute calls for [\w\.]+: \d+")
    ),
    (
        "other message with activation",
        dict(activate=True),
        False,
        None,
    ),
    (
        "second message with deactivation",
        dict(activate=False),
        False, 
        dict(levelno=LogLevel.FULL_DEBUG, pattern=r"Time steps:\n")
    ),
    ("common message", dict(), True, None),
])
def test_ExplicitTimeDriver_log_debug_message(format, msg, kwargs, to_log, emitted):
    # Attribute of a mock object can be created by passing them as kwargs.
    # Here, we get an object in the variable handler with two members:
    # an attribute level == LogLevel.DEBUG and an mocked attribute log
    handler = MagicMock(level=LogLevel.DEBUG, log=MagicMock())
    rec = logging.getLogRecordFactory()("log_test", LogLevel.INFO, __file__, 22, msg, (), None)
    for key, value in kwargs.items():
        setattr(rec, key, value)

    d = ExplicitTimeDriver("dummy")

    assert d.log_debug_message(handler, rec, format) == to_log

    if "activate" in kwargs and not msg.endswith("_run"):
        assert d.record_dt == kwargs["activate"]

    if emitted:
        handler.log.assert_called_once()
        args = handler.log.call_args[0]
        assert args[0] == emitted["levelno"]
        assert re.match(emitted["pattern"], args[1]) is not None
    else:
        handler.log.assert_not_called()


def test_ExplicitTimeDriver_init_conditions():
    """
    Test checking that initial conditions are applied before any system compute.
    Related to https://gitlab.com/cosapp/cosapp/-/issues/20
    """
    class Dummy(System):
        def setup(self):
            self.add_inward('c', 1.0)
            self.add_transient('x', der='c - c')

        def compute(self):
            if self.x <= 0:
                raise RuntimeError("x must be positive")
            if self.c >= 0:
                raise RuntimeError("c must be negative")

    system = Dummy('dummy')
    driver = system.add_driver(ExplicitTimeDriver(dt=0.1, time_interval=[0, 0.1]))

    driver.set_scenario(
        init = {'x': 1.0},  # valid initial value for x
        values = {'c': '-(1 + t**2)'},  # valid c(t) value
    )

    system.x = system.c = -1.0
    with pytest.raises(RuntimeError, match="x must be positive"):
        system.run_once()

    system.x = system.c = 1.0
    with pytest.raises(RuntimeError, match="c must be negative"):
        system.run_once()

    system.c, system.x = 1.0, -1.0  # both invalid
    with pytest.raises(RuntimeError):
        system.run_once()
    
    # Running the time driver should not raise RuntimeError,
    # as both x and c are expected to be set at valid values
    with not_raised(RuntimeError):
        system.run_drivers()

    assert system.x == 1
    assert system.c == pytest.approx(-1.01)



"""
Core package of CoSApp.
"""
from cosapp.core._version import __version__

from cosapp.core.numerics.basics import MathematicalProblem
from cosapp.core.numerics.root import root
from cosapp.core.config import CoSAppConfiguration

from cosapp.core.module import Module

__all__ = ["MathematicalProblem", "root"]



"""
Classes connecting `Port` of foreign `System` to transfer variable values.
"""
import copy
import logging
import weakref
from typing import Callable, Dict, List, Optional, Tuple, Union

from cosapp.ports import units
from cosapp.ports.port import ExtensiblePort, PortType
from cosapp.utils.helpers import check_arg, is_numerical

logger = logging.getLogger(__name__)


class ConnectorError(Exception):
    """Raised if a connector cannot be created between two `Port`.

    Attributes
    ----------
    message : str
        Error message
    """

    def __init__(self, message: str):
        """Instantiate a error object from the error descriptive message.

        Parameters
        ----------
        message : str
            Error message
        """
        self.message = message


class Connector:
    """This class connect two ports without enforcing that all port variables are connected.

    The link is oriented from the source to the sink.

    Parameters
    ----------
    name : str
        Name of the connector
    source : ExtensiblePort
        Port from which originate the variables
    mapping : str or List[str] or Dict[str, str]
        (List of) common name(s) or mapping name dictionary
    sink : ExtensiblePort
        Port to which the variables are transferred
    """

    def __init__(
        self,
        name: str,
        sink: ExtensiblePort,
        source: ExtensiblePort,
        mapping: Union[str, List[str], Dict[str, str], None] = None,
    ):
        """Connector constructor from the two `ExtensiblePort` to link and the list of variables to map.

        If no mapping is provided, connection will be made between variables based on their names. 
        If a name mapping is provided as a list, the name should be present in both port. And if 
        the mapping is specified as a dictionary, the keys belong to the sink port and the values
        to the source port.

        Parameters
        ----------
        name : str
            Name of the connector
        sink : ExtensiblePort
            Port to which the variables are transferred.
        source : ExtensiblePort
            Port from which originate the variables.
        mapping : str or List[str] or Dict[str, str], optional
            (List of) common name(s) or mapping name dictionary; default None (i.e. no mapping).
        
        Raises
        ------
        ConnectorError
            If the connection between the `source` and the `sink` is not possible to establish.
        """
        self._name = name  # type: str

        if not source.owner:
            raise ConnectorError("Source owner is undefined.")

        if not sink.owner:
            raise ConnectorError("Sink owner is undefined.")

        if source is sink:
            raise ConnectorError("Source and sink cannot be the same object.")

        self._source = weakref.ref(source)  # type: ExtensiblePort
        self._sink = weakref.ref(sink)  # type: ExtensiblePort

        # Generate mapping dictionary
        if mapping is None:
            mapping = dict([(v, v) for v in sink if v in source])
        elif isinstance(mapping, str):
            mapping = dict([(mapping, mapping)])
        elif not isinstance(mapping, dict):
            mapping = dict(zip(mapping, mapping))

        self._mapping = mapping  # type: Dict[str, str]

        # Validate the connection based on variable attributes
        for target, origin in self._mapping.items():
            if origin not in self.source:
                raise ConnectorError(
                    f"Source port {self.source} has no variable '{origin}'"
                )
            if target not in self.sink:
                raise ConnectorError(
                    f"Sink port {self.source} has no variable '{target}'"
                )

        self._unit_conversions = dict(
            [(name, None) for name in self._mapping]
        )  # type: Dict[str, Optional[Tuple[float, float]]]

        self.update_unit_conversion()

    def __repr__(self) -> str:
        return "Connector({} <- {}, {})".format(
            self.sink.contextual_name,
            self.source.contextual_name,
            self.variable_mapping,
        )

    @property
    def name(self) -> str:
        """str : name of the connector."""
        return self._name

    @property
    def source(self) -> ExtensiblePort:
        """`ExtensiblePort`: Port from which transferred values originate."""
        return self._source()

    @source.setter
    def source(self, port: ExtensiblePort) -> None:
        check_arg(port, 'source', ExtensiblePort)

        if len(self.source) != len(port):
            raise TypeError(
                f"New port {port} has not the same variables as the source {self.source}."
            )

        for variable in self.source:
            if variable not in port:
                raise AttributeError(
                    f"Source variable {variable} does not exist in port {port}."
                )

        self._source = weakref.ref(port)

    @property
    def sink(self) -> ExtensiblePort:
        """`ExtensiblePort`: Port to which values are transferred."""
        return self._sink()

    @sink.setter
    def sink(self, port: ExtensiblePort) -> None:
        check_arg(port, 'sink', ExtensiblePort)

        if len(self.sink) != len(port):
            raise TypeError(
                f"New port {port} has not the same variables as the sink {self.sink}."
            )

        for variable in self.sink:
            if variable not in port:
                raise AttributeError(
                    f"Sink variable {variable} does not exist in port {port}."
                )

        self._sink = weakref.ref(port)

    @property
    def variable_mapping(self) -> Dict[str, str]:
        """Dict[str, str] : Variable name mapping between the sink (key) and the source (value)."""
        return self._mapping

    def set_perturbation(self, name: str, value: float) -> None:
        """Add a perturbation on a connector.
        
        Parameters
        ----------
        name : str
            Name of the sink variable to perturb
        value : float
            Perturbation value
        """
        if self._unit_conversions[name] is None:
            ValueError(f"'{name}' does not have a numerical type.")

        s, a = self._unit_conversions[name]
        self._unit_conversions[name] = s, a + value
        self.source.owner.set_dirty(PortType.IN)

    def remove_variables(self, names: List[str]) -> None:
        """Remove the provided variables from this connection.

        The provided names should be sink names.

        Parameters
        ----------
        names : List[str]
            list of sink variable names to be removed.
        """
        for variable in names:
            del self._mapping[variable]
            del self._unit_conversions[variable]

    def update_unit_conversion(self, name: Optional[str] = None) -> None:
        """Update the physical unit conversion on the connector.

        If `name` is not `None`, update the conversion only for the connexion towards that variable.

        Parameters
        ----------
        name : str, optional
            Name of the variable for which unit conversion needs an update; default None (i.e. all
            conversions will be updated).

        Raises
        ------
        UnitError
            If unit conversion from source to sink is not possible
        """

        def update_one_connection(key: str) -> None:
            """Update the unit converter of the connected key.

            Parameters
            ----------
            key : str
                Name of the connected variable for which the unit converter should be updated.
            """
            source_name = self.variable_mapping[key]
            source_unit = self.source.get_details(source_name).unit
            target_unit = self.sink.get_details(key).unit
            # Get the conversion between unit and check it is valid
            converter = units.get_conversion(source_unit, target_unit)
            if converter is None:
                # Print a warning if one hand of the connexion is a valid unit but not the other one
                if units.is_valid_units(source_unit):
                    logger.warning(
                        f"Connector source has physical unit {source_unit}, but the sink is dimensionless."
                    )
                elif units.is_valid_units(target_unit):
                    logger.warning(
                        f"Connector source is dimensionless, but the sink has physical unit {target_unit}."
                    )

                if is_numerical(self.sink[key]):
                    converter = (1.0, 0.0)

            self._unit_conversions[key] = converter

        if name is None:
            for name in self._unit_conversions:
                update_one_connection(name)
        else:
            update_one_connection(name)

    def transfer(self) -> None:
        """Transfer values from `source` to `sink`."""
        # TODO improve efficiency
        def conversion_function(key) -> Callable:
            try:
                slope, offset = self._unit_conversions[key]
            except:
                converter = copy.deepcopy
            else:
                converter = lambda var: slope * (var + offset)
            return converter

        source, sink = self.source, self.sink

        if not source.owner.is_clean(source.direction):
            sink.owner.set_dirty(sink.direction)

            for key, item in self._mapping.items():
                # get/setattr faster for Port
                target = getattr(source, item)
                convert = conversion_function(key)
                try:
                    setattr(sink, key, convert(target))
                except TypeError:
                    setattr(sink, key, copy.deepcopy(target))

    def to_dict(self) -> Dict[str, Union[str, Tuple[Dict[str, str], str]]]:
        """Convert this connector to a dictionary.

        The dictionary returned has the following structure
        Dict[str, Tuple[Dict[str, str], str]]::

           {
             port_name: (variable mapping, source name)
           }

        Or if the mapping is complete with the same name::

           {
             port_name: source name
           }

        Returns
        -------
        dict
            Dictionary representing this connector
        """
        tmp = dict()

        target = self.sink.contextual_name
        if self.sink.owner is self.source.owner.parent:
            target = self.sink.name
        origin = self.source.contextual_name
        if self.source.owner is self.sink.owner.parent:
            origin = self.source.name

        # If the mapping is full and with the same nomenclature
        same_nomenclature = False
        if len(self._mapping) == len(self.source) == len(self.sink):
            same_nomenclature = True
            for k, v in self._mapping.items():
                if k != v:
                    same_nomenclature = False
                    break

        if same_nomenclature:
            tmp[self.name] = (target, origin)
        else:
            tmp[self.name] = (target, origin, self._mapping.copy())

        return tmp



"""Recorder to Delimited Separated Value format file."""
import os
from typing import Any, List, Optional, Union
from collections.abc import Collection

import numpy
import pandas

from cosapp.recorders.recorder import BaseRecorder
from cosapp.utils.helpers import is_numerical, check_arg


class DSVRecorder(BaseRecorder):
    """Record data into Delimiter Separated Value file.

    Matching pattern are case sensitive and support the following special patterns:

    ========  ================================
    Pattern   Meaning
    ========  ================================
    `*`       matches everything
    `?`       matches any single character
    `[seq]`   matches any character in seq
    `[!seq]`  matches any character not in seq
    ========  ================================

    Excluding pattern are shadowing includes one; e.g. if `includes='*port_in.*' and `excludes='*.Pt'`, for a port
    having variables named `Tt` and `Pt`, only `Tt` will be recorded.

    Parameters
    ----------
    filepath : str
        Filepath to save data into.
    delimiter : `','`, `';'` or `'\t'`, optional
        Delimiter of data in the file; default `','`.
    buffer : bool, optional
        Should the data written after the simulation (`False`) or every time they are available (`True`);
        default `False`.
    includes : str or list of str, optional
        Variables matching these patterns will be included; default `'*'` (i.e. all variables).
    excludes : str or list of str or None, optional
        Variables matching these patterns will be excluded; default `None` (i.e. nothing is excluded).
    numerical_only : bool, optional
        Keep only numerical variables (i.e. number or numerical vector); default False.
    section : str, optional
        Current section name; default `''`.
    precision : int, optional
        Precision digits when writing floating point number; default 9 (i.e. 10 figures will be written).
    hold : bool, optional
        Append the new data or not; default `False`.
    raw_output : bool, optional
        Raw output; default `False`.

    ..note::
        Do not mention `inwards` or `outwards` in `includes` or `excludes` list. Otherwise you may not record the wanted
        variables.
    """

    def __init__(
        self,
        filepath: str,
        delimiter = ",",
        buffer = False,
        includes: Union[str, List[str]] = "*",
        excludes: Optional[Union[str, List[str]]] = None,
        numerical_only = False,
        section = "",
        precision = 9,
        hold = False,
        raw_output = False,
    ):
        check_arg(filepath, "filepath", str)
        check_arg(delimiter, "delimiter", str)
        check_arg(buffer, "buffer", bool)

        supported_delimiters = (",", ";", "\t")
        if delimiter not in supported_delimiters:
            raise ValueError(
                f"Supported delimiters are {supported_delimiters}; got {delimiter!r}"
            )

        super().__init__(
            includes, excludes, numerical_only, section, precision, hold, raw_output
        )

        self.__filepath = filepath  # type: str
        self.__delimiter = delimiter  # type: str
        self.__buffer = list() if buffer else None  # type: List[List[Any]]

    @property
    def filepath(self) -> str:
        """str: path of the DSV file"""
        return self.__filepath

    @property
    def delimiter(self) -> str:
        """str: column delimiter used in DSV file"""
        return self.__delimiter

    def export_data(self) -> pandas.DataFrame:
        """Export recorded results into a pandas.DataFrame object."""
        if not os.path.exists(self.__filepath):
            return pandas.DataFrame()
        elif self.__buffer is not None:
            with open(self.__filepath, "r") as f:
                headers = f.readline().split(self.__delimiter)
            return pandas.DataFrame(self.__buffer, columns=headers)
        else:
            return pandas.read_csv(self.__filepath, delimiter=self.__delimiter, header=0)

    @property
    def _raw_data(self) -> List[List[Any]]:
        """Return a raw/unformatted version of records.

        Returns
        -------
        List[List[Any]]
            Records of `watched_object` for variables given by method `field_names()`
        """
        if not os.path.exists(self.__filepath):
            return list()

        if self.__buffer is not None:
            return self.__buffer

        with open(self.__filepath, "r") as f:
            # The header line is skipped
            content = map(lambda line: line.split(self.__delimiter), f.readlines()[1:])
        return content

    def start(self):
        """Initialize recording support."""
        super().start()
        if not self.hold:
            # TODO Fred we could use clean/dirty here
            self.watched_object.run_once()  # Run the System to be sure the list are developed.
            if self.__buffer is not None:
                self.__buffer.clear()

            headers = [
                self.SPECIALS.section,
                self.SPECIALS.status,
                self.SPECIALS.code,
                self.SPECIALS.reference,
            ]

            def add_header(base, unit):
                if unit and not self._raw_output:
                    headers.append(f"{base} [{unit}]")
                else:
                    headers.append(base)

            varlist = self.field_names()
            for name, unit in zip(varlist, self._get_units(varlist)):
                value = self.watched_object[name]
                unit = "" if self._raw_output else unit
                if isinstance(value, numpy.ndarray) and value.ndim > 0:
                    for i in range(value.size):
                        entry = f"{name}[{i}]"
                        add_header(entry, unit)
                elif isinstance(value, Collection) and not isinstance(value, str):
                    for i in range(len(value)):
                        entry = f"{name}[{i}]"
                        add_header(entry, unit)
                else:
                    add_header(name, unit)

            with open(self.__filepath, "w") as f:
                # Write header
                f.write(self.__delimiter.join(headers) + "\n")

    def formatted_data(self) -> List[Any]:
        """Format collected data from watched object into a list."""

        precision = self.precision
        def fmt(value, check=True):
            if check and not is_numerical(value):
                return str(value)
            return "{0:.{1}e}".format(value, precision)

        line = []
        for value in self.collected_data():
            if isinstance(value, numpy.ndarray):
                if numpy.issubdtype(value.dtype, numpy.number):
                    line.extend(fmt(v, check=False) for v in value.flat)
                else:
                    line.extend(str(v) for v in value.flat)
            elif isinstance(value, Collection) and not isinstance(value, str):
                line.extend(fmt(v) for v in value)
            else:
                line.append(fmt(value))

        return line

    def _record(self, line: List[Any]) -> None:
        if self.__buffer is None:
            with open(self.__filepath, "a") as f:
                f.write(self.__delimiter.join(line) + "\n")
        else:
            self.__buffer.append(line)

    def clear(self):
        """Clear all previously stored data."""
        if self.__buffer is not None:
            self.__buffer.clear()
        super().clear()

    def exit(self):
        """Close recording session."""
        if self.__buffer is not None:
            with open(self.__filepath, "a") as f:
                for line in self.__buffer:
                    f.write(self.__delimiter.join(line) + "\n")



"""Basic class to define a variable distribution."""
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

import scipy
import numpy

EPS = 1e-12


class Distribution(ABC):
    """Abstract class to define a variable distribution.

    Parameters
    ----------
    worst : float
        The parameter absolute worst value.
    best : float
        The parameter absolute best value.
    pworst : float, optional
        The worst value probability is the probability that the variable will be lower (if worst<best) or higher
        (if worst>best) than the worst value; default 0.15 (i.e. 15%).
    pbest : float, optional
        The best value probability is the probability that the variable will be higher (if worst<best) or lower
        (if worst>best) than the best value; default 0.15 (i.e. 15%).
    """

    # TODO should we forbid worst > best

    __slots__ = ("__wearkref__", "_best", "_pbest", "_pworst", "_worst")

    def __init__(
        self,
        worst: float,
        best: float,
        pworst: Optional[float] = 0.15,
        pbest: Optional[float] = 0.15,
    ):
        self._worst = worst
        self._best = best
        # Need to initiate with EPS to be sure to fulfill pworst + pbest < 1.
        # We don't initialise to 0. otherwise fitting distribution for unbounded
        # one will fail (like the normal distribution)
        self._pworst = EPS
        self._pbest = EPS

        self.pworst = pworst
        self.pbest = pbest

    def __json__(self) -> Dict[str, Any]:
        """Serialize the distribution object.

        Returns
        -------
        Dict[str, Any]
            JSONable dictionary describing the distribution.
        """
        return {
            "worst": self.worst,
            "pworst": self.pworst,
            "best": self.best,
            "pbest": self.pbest,
        }

    @abstractmethod
    def _set_distribution(self) -> None:
        """Set the probability distribution according the parameters."""
        pass

    @property
    def worst(self) -> float:
        """float: The worst possible value of the variable.
        
        It is defined as an absolute value. It has a physical meaning, i.e. it
        can be numerically higher than the best value.
        The probability to meet the worst value is given by the `pworst` property.
        """
        return self._worst

    @worst.setter
    def worst(self, value):
        self._worst = value
        self._set_distribution()

    @property
    def pworst(self) -> float:
        """float: The worst value probability.
        
        It is the probability that the variable will be lower (if worst<best) or higher
        (if worst>best) than the worst value.
        """
        return self._pworst

    @pworst.setter
    def pworst(self, value: float):
        if not (0.0 <= value <= 1.0):
            raise ValueError(
                f"Worst probability does not verify 0 <= {value} <= 1."
            )
        self._pworst = value
        self._set_distribution()

    @property
    def best(self) -> float:
        """float : The best possible value of the variable.
        
        It is defined as an absolute value. It has a physical meaning, i.e. it can
        be numerically lower than the worst value.
        The probability to meet the best value is given by the `pbest` property.
        """
        return self._best

    @best.setter
    def best(self, value: float):
        self._best = value
        self._set_distribution()

    @property
    def pbest(self) -> float:
        """float: The best value probability.
        
        It is the probability that the variable will be higher (if worst<best) or lower
        (if worst>best) than the best value.
        """
        return self._pbest

    @pbest.setter
    def pbest(self, value: float):
        if not (0.0 <= value <= 1.0):
            raise ValueError(
                f"Best probability does not verify 0 <= {value} <= 1."
            )
        self._pbest = value
        self._set_distribution()

    @abstractmethod
    def draw(self, quantile: Optional[float] = None) -> float:
        """Generate a random number.

        If a quantile is given, generate the perturbation for that quantile.

        Parameters
        ----------
        quantile : Optional[float], optional
            Quantile for which the perturbation must be set; default None (i.e. random perturbation)
        
        Returns
        -------
        float
            The random number
        """
        pass



from enum import Enum

# TODO
# class SolverStep(Enum):
#     """Step in progress within the numerical solver.
#
#     INIT : Initialization phase
#     JACOBIAN : Jacobian matrix building phase
#     ITERATION : Convergence iteration
#     """
#     INIT = 'init'
#     JACOBIAN = 'jacobian'
#     ITERATION = 'iteration'


class NonLinearMethods(Enum):
    """Enumeration of non-linear algorithm available.

    POWELL : Modified Powell method using MINPACK’s hybrd and hybrj routines
    BROYDEN_GOOD : Broyden’s first Jacobian approximation. This method is also known as “Broyden’s good method”.
    NR : Simple Newton-Raphson
    """

    POWELL = "hybr"
    BROYDEN_GOOD = "broyden1"
    NR = "cosapp"




import pytest

import numpy as np
from cosapp.drivers import EulerExplicit, NonLinearSolver, RunSingleCase
import cosapp.recorders as recorders


def test_EulerExplicit_init_default():
    driver = EulerExplicit()
    assert driver.owner is None
    assert driver.dt is None
    assert driver.time_interval is None
    assert driver.name == "Explicit Euler time driver"


def test_EulerExplicit_ode_solve_1(ode_case_1):
    # First run
    ode, driver = ode_case_1(EulerExplicit, dt=1e-3, time_interval=[0, 1])

    ode.run_drivers()
    start, end = driver.time_interval
    assert ode.f == pytest.approx(ode(end), 1e-3)
    assert end == 1

    # Second run, where end time is not a multiple of time step
    ode, driver = ode_case_1(EulerExplicit, dt=1e-3, time_interval=[0, 1 + 2e-4])
    ode.run_drivers()
    start, end = driver.time_interval
    assert ode.f == pytest.approx(ode(end), 1e-3)
    assert end == pytest.approx(1.0002)


@pytest.mark.parametrize("dt", [
    0.1, 0.01,
    # 1e-3,  # long
])
def test_EulerExplicit_twoTanks(two_tank_case, two_tank_solution, dt):
    system, driver = two_tank_case(EulerExplicit, dt=dt, time_interval=[0, 5])
    solver = driver.add_child(NonLinearSolver('solver', factor=1.0))

    h1_0, h2_0 = init = (3, 1)

    driver.set_scenario(
        name = 'run',
        init = {
            'tank1.height': h1_0,  # initial conditions
            'tank2.height': h2_0,
        },
        values = {
            'pipe.D': 0.07,  # fixed values
            'pipe.L': 2.5,
            'tank1.area': 2,
        }
    )
    
    assert driver.scenario.context is system
    assert driver.scenario.name == 'run'
    
    recorder = driver.add_recorder(recorders.DataFrameRecorder(includes='tank?.height'), period=0.1)
    # recorder = driver.add_recorder(recorders.DSVRecorder('twoTanks_Euler.csv', includes=['tank?.height', ]), period=0.1)
    assert driver.recording_period == 0.1

    system.run_drivers()

    assert system.tank1.height < 3
    assert system.tank2.height > 1
    assert system.tank1.height == pytest.approx(system.tank2.height, rel=1e-3)

    df = recorder.export_data()
    assert len(df) == 51
    solution = two_tank_solution(system, init)
    assert solution.characteristic_time == pytest.approx(0.5766040318109212)
    time = np.array(df['Reference'], dtype=float)
    error = 0
    for t, h1 in zip(time, df['tank1.height']):
        exact = solution(t)
        error = max(error, abs(h1 - exact[0]))
    # Test that maximum error ~ dt
    assert error < 0.3 * dt
    assert error > 0.2 * dt



"""
Classes driving simulation on CoSApp :py:class:`~cosapp.systems.system.System`.
"""
import logging
import re
import time
from collections import OrderedDict
from typing import List, Dict, Optional, Any

from cosapp.core.module import Module
from cosapp.recorders.recorder import BaseRecorder
from cosapp.utils.options_dictionary import OptionsDictionary
from cosapp.utils.naming import NameChecker
from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


class Driver(Module):
    """Abstract base class for all systems drivers.

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
    **kwargs : Any
        Keyword arguments will be used to set driver options

    Attributes
    ----------
    name : str
        Name of the driver
    parent : Driver, optional
        Top driver containing this driver; default None.
    owner : System
        :py:class:`~cosapp.systems.system.System` to which this driver belong.
    children : OrderedDict[str, Driver]
        Drivers belonging to this one.

    options : OptionsDictionary
      |  Options for the current driver:
      |  **verbose** : int, {0, 1}
      |      Verbosity level of the driver; default 0 (i.e. minimal information)

    solution :  List[Tuple[str, float]]
        List of (name, value) for the iteratives when a solution is reached
    """

    __slots__ = ('_owner', '_recorder', 'options', 'start_time', 'status', 'error_code')

    _name_check = NameChecker(
        pattern = r"^[A-Za-z][\w\s@-]*[\w]?$",
        message = "Driver name must start with a letter, and contain only alphanumerics + {'_', '@', ' ', '-'}",
        excluded = ["inwards", "outwards"],
    )

    def __init__(self,
        name: str,
        owner: "Optional[cosapp.systems.System]" = None,
        **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name)
        self._owner = None  # type: Optional[System]
        self._recorder = None  # type: Optional[BaseRecorder]
        self.owner = owner

        self.options = OptionsDictionary()  # type: OptionsDictionary
            # "Driver options dictionary"
        self.start_time = 0.0  # type: float
            # unit="s",
            # desc="Absolute time at which the Driver execution started.",
        self.status = ""  # type: str
            #desc="Status of the driver."
            # TODO Fred what are the status? Enum, any str?
        self.error_code = "0"  # type: str
            # desc="Error code during the execution."
            # TODO Fred what is the code? ESI?

        # TODO is tol_target really used in all cases? Is it not redundant with options parameters?
        # self.target_tol = None # desc='Targeted maximal relative tolerance for numerical solution.')
        # TODO is current_tol updated in all cases to a relevant value?
        # self.current_tol = np.nan # desc='Maximal current relative tolerance of the numerical solution.')

        self.options.declare(
            "verbose",
            default=0,
            dtype=int,
            lower=0,
            upper=1,
            desc="Verbosity level of the driver",
        )
        for key in self.options:
            try:
                self.options[key] = kwargs.pop(key)
            except KeyError:
                continue

    def __repr__(self) -> str:
        context = "alone" if self.owner is None else f"on System {self.owner.name!r}"
        return f"{self.name} ({context}) - {self.__class__.__name__}"

    @property
    def owner(self) -> "Optional[cosapp.systems.System]":
        """System : System owning the driver and its children."""
        return self._owner

    @owner.setter
    def owner(self, system: "Optional[cosapp.systems.System]") -> None:
        from cosapp.systems import System
        if system is not None:
            check_arg(system, 'owner', System)

        self._owner = system
        if self._recorder is not None:
            self._recorder.watched_object = self._owner
        for child in self.children.values():
            child.owner = system

    def check_owner_attr(self, item: str) -> None:
        if item not in self.owner:
            raise AttributeError(f"{item!r} not found in System {self.owner.name!r}")

    @property
    def recorder(self) -> Optional[BaseRecorder]:
        """BaseRecorder or None : Recorder attached to this `Driver`."""
        return self._recorder

    def is_standalone(self) -> bool:
        """Is this Driver able to solve a system?

        Returns
        -------
        bool
            Ability to solve a system or not.
        """
        for driver in self.children.values():
            if driver.is_standalone():
                return True

        return False

    def _set_children_active_status(self, active_status : bool) -> None:
        self._active = active_status
        for child in self.children.values():
            child._set_children_active_status(active_status)

    def _precompute(self) -> None:
        """Set execution order and start the recorder."""
        if self.owner is None:
            raise AttributeError(f"Driver {self.name!r} has no owner system.")

        if self._recorder is not None:
            self._recorder.start()

        # Set execution order as Driver appending order
        if len(self.exec_order) != len(self.children):
            self.exec_order = list(self.children)

        self.start_time = time.time()

        if self.owner.parent is None and self.parent is None:
            logger.info(" " + "-" * 60)
            logger.info(f" # Starting driver {self.name!r} on {self.owner.name!r}")

    def _postcompute(self) -> None:
        """Actions performed after the `Module.compute` call."""
        if self._recorder is not None:
            self._recorder.exit()  # TODO Fred Better in clean_run ?

        if self.owner.parent is None and self.parent is None:
            logger.info(
                " # Ending driver {!r} on {!r} in {} seconds\n".format(
                    self.name, self.owner.name, round(time.time() - self.start_time, 3)
                )
            )

    def add_child(
        self,
        child: "Driver",
        execution_index: Optional[int] = None,
    ) -> "Driver":
        """Add a child `Driver` to the current `Driver`.

        When adding a child `Driver`, it is possible to specified its position in the execution
        order.

        Parameters
        ----------
        child: Driver
            `Driver` to add to the current `Driver`
        execution_index: int, optional
            Index of the execution order list at which the `Module` should be inserted;
            default latest.

        Notes
        -----
        The added child will have its owner set to match the one of the current driver.
        """
        check_arg(child, 'child', Driver)
        
        driver = super().add_child(child, execution_index)
        driver.owner = self.owner

        return driver

    def add_driver(
        self,
        child: "Driver",
        execution_index: Optional[int] = None,
    ) -> "Driver":
        """Alias for :py:meth:`~cosapp.drivers.driver.Driver.add_child`."""
        return self.add_child(child, execution_index)

    def add_recorder(self, recorder: BaseRecorder) -> BaseRecorder:
        check_arg(recorder, 'recorder', BaseRecorder)

        self._recorder = recorder
        self._recorder.watched_object = self.owner
        self._recorder._owner = self

        return self.recorder



import abc
import json
import logging
from copy import copy
from numbers import Number
from typing import (Any, AnyStr, Callable, Dict, List, Optional,
                    Sequence, Tuple, Union)

import numpy

from cosapp.core.numerics.basics import MathematicalProblem, SolverResults
from cosapp.drivers.driver import Driver
from cosapp.drivers.iterativecase import IterativeCase
from cosapp.drivers.runonce import RunOnce
from cosapp.utils.options_dictionary import OptionsDictionary

logger = logging.getLogger(__name__)


class AbstractSolver(Driver):
    """
    Solve a `System`

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
    **kwargs : Any
        Keyword arguments will be used to set driver options
    """

    __slots__ = ('force_init', 'problem', 'initial_values', 'solution')

    def __init__(self,
        name: str,
        owner: "Optional[cosapp.systems.System]" = None,
        **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)

        self.force_init = False  # type: bool
            # desc="Force the initial values to be used."

        self.problem = None  # type: Optional[MathematicalProblem]
            # desc="Mathematical problem to solve."
        self.initial_values = numpy.empty(0, dtype=float)  # type: numpy.ndarray
            # desc="List of initial values for all iteratives.",
        self.solution = {}  # type: Dict[str, float]
            # desc="Dictionary (name, value) of the latest solution reached."

    def _filter_options(self, kwargs, aliases: Dict[str, str] = dict()):
        """
        Translate option names into self.options using an alias dictionary, to handle cases where
        a common option name, such as 'tol', is passed to a specific solver/function with a different name.
        
        For example, in scipy.optimize.root(), the convergence criterion may be referred to as 'ftol', 'gtol'...
        depending on the invoked algorithm (Levenberg-Marquardt, Powell, Broyden's good, etc.).
        """
        keys = list(kwargs.keys())  # make a copy of keys only

        for key in keys:
            value = kwargs.pop(key)
            try:
                key = aliases[key]
            except KeyError:
                pass
            if key in self.options:
                self.options[key] = value
            else:
                raise KeyError(f"Unknown solver option {key!r}")

    @property
    def _default_driver_name(self) -> str:
        """str : Name for the default driver."""
        return "runner"

    def _get_solver_limits(self) -> Dict[str, numpy.ndarray]:
        """Returns the step limitations for all iteratives.

        There are 4 types of limits defined:
        - lower_bound: lower bound of the iteratives
        - upper_bound: upper bound of the iteratives
        - abs_step: maximal absolute step of the iteratives
        - rel_step: maximal relative step of the iteratives

        Returns
        -------
        Dict[str, numpy.ndarray]
            Dictionary with the limits for all iteratives.
        """
        options = {
            name : []
            for name in ("lower_bound", "upper_bound", "abs_step", "rel_step")
        }

        for unknown in self.problem.unknowns.values():
            def append(name, const):
                array = numpy.full_like(unknown.value, const, dtype=type(const))
                options[name] = numpy.concatenate((options[name], array.flatten()))
            append("lower_bound", unknown.lower_bound)
            append("upper_bound", unknown.upper_bound)
            append("abs_step", unknown.max_abs_step)
            append("rel_step", unknown.max_rel_step)

        return options

    def compute_before(self):
        """Contains the customized `Module` calculation, to execute before children.
        """
        logger.debug(f"Set unknowns initial values: {self.initial_values}")
        x = self.initial_values
        counter = 0
        # Pass the x vector to all driver so they set the new x (at appropriate time)
        for child in self.children.values():
            if isinstance(child, IterativeCase):
                counter += child.set_iteratives(x[counter:])

    def _precompute(self) -> None:
        """Set up the mathematical problem."""
        # TODO we should check that all variables are of numerical types
        super()._precompute()

        self.initial_values = numpy.empty(0, dtype=float)
        self.problem = MathematicalProblem(self.name, self.owner)

    def _fresidues(self,
        x: Sequence[float],
        update_residues_ref: bool = True,
    ) -> numpy.ndarray:
        """
        Method used by the solver to take free variables values as input and values of residues as
        output (after running the System).

        Parameters
        ----------
        x : Sequence[float]
            The list of values to set to the free variables of the `System`
        update_residues_ref : bool
            Request residues to update their reference

        Returns
        -------
        numpy.ndarray
            The list of residues of the `System`
        """
        x = numpy.asarray(x)
        logger.debug(f"Call fresidues with x = {x!r}")
        counter = 0
        # Pass the x vector to all driver so they set the new x (at appropriate time)
        for child in self.children.values():
            if isinstance(child, IterativeCase):
                counter += child.set_iteratives(x[counter:])

        # Run all points
        for child in self.exec_order:
            logger.debug(f"Call {child}.run_once")
            self.children[child].run_once()

        residues = self.problem.residues_vector
        logger.debug(f"Residues: {residues!r}")
        return residues

    @abc.abstractmethod
    def resolution_method(self,
        fresidues: Callable[[Sequence[float], Union[float, str], bool], numpy.ndarray],
        x0: Sequence[float],
        args: Tuple[Union[float, str]] = (),
        options: Optional[OptionsDictionary] = None,
    ) -> SolverResults:
        """Function call to cancel the residues.

        Parameters
        ----------
        fresidues : Callable[[Sequence[float], Union[float, str]], numpy.ndarray]
            Residues function taking two parameters (evaluation vector, time/ref) and returning the residues
        x0 : Sequence[float]
            The initial values vector to converge to the solution
        args : Tuple[Union[float, str], bool], optional
            A tuple of additional argument for fresidues starting with the time/ref parameter and the need to update
            residues reference
        options : OptionsDictionary, optional
            Options for the numerical resolution method

        Returns
        -------
        SolverResults
            Solution container
        """
        pass

    # Don't clean initial_values and problem => could be useful for debugging
    # def _postcompute(self) -> None:
    #     """Undo pull inputs and reset iteratives sets."""
    #     self.initial_values = numpy.empty(0, dtype=float)
    #     self.problem = None
    #     super()._postcompute()

    def save_solution(self, file: Optional[str] = None) -> Dict[str, Union[Number, List[Number]]]:
        """Save the latest solver solution.

        If `file` is specified, the solution will be saved in it in JSON format.

        Parameters
        ----------
        file : str, optional
            Filename to save the answer in; default None (i.e. data will not be saved)

        Returns
        -------
        Dict[str, Union[Number, List[Number]]]
            Dictionary of the latest solution
        """
        latest_answer = dict()

        for k, v in self.solution.items():
            if isinstance(v, numpy.ndarray):
                v = v.tolist()
            latest_answer[k] = copy(v)

        if file:
            with open(file, "w") as outfile:
                json.dump(latest_answer, outfile)

        return latest_answer

    def load_solution(self,
        solution: Union[Dict[str, Union[Number, numpy.ndarray]], AnyStr],
        case: Optional[str] = None,
    ):
        """Load the provided solution to initialize the solver.

        The solution can be provided directly as a dictionary or from a filename to be read.

        Parameters
        ----------
        solution : Dict[str, Union[Number, numpy.ndarray]] or str
            Dictionary of the latest solution to load or the filename in JSON format to read from.
        case : str, optional
            Case to initialize with the solution; default None (i.e. will be guessed from variable name)
        """
        # TODO Fred is it better to set the initial values or to override the previous solution?
        #   In case the solution does not cover all offdesign unknowns, the later should be better.
        from cosapp.systems import System

        if not isinstance(solution, (dict, str)):
            raise TypeError(
                "Answer expected as dict or str name of the json file, got {}".format(
                    type(solution).__qualname__
                )
            )

        if isinstance(solution, str):
            with open(solution, "r") as f:
                data = json.load(f)
        else:
            data = solution

        with System.set_master(repr(self.owner)) as is_master:
            if is_master:
                self.owner.open_loops()

            try:
                if case is None:
                    for name, value in data.items():
                        for child in filter(
                            lambda d: isinstance(d, RunOnce), self.children.values()
                        ):
                            if name.startswith(child.name):  # We got a offdesign variable
                                name = name[len(child.name) + 1 : -1]
                                if name in self.owner:
                                    child.set_init({name: numpy.asarray(value)})
                                    break
                            elif (
                                name in child.design.unknowns
                            ):  # We may have a design variable
                                child.set_init({name: numpy.asarray(value)})
                                break
                else:
                    child = self.children[case]
                    if not isinstance(child, RunOnce):
                        raise TypeError(
                            "Only driver derived from RunOnce can be initialized; got '{}' for driver '{}'.".format(
                                type(child).__qualname__, case
                            )
                        )

                    for name, value in data.items():
                        if name.startswith(child.name):
                            name = name[len(child.name) + 1 : -1]
                        if name in self.owner:
                            child.set_init({name: numpy.asarray(value)})

            finally:  # Ensure to clean the system
                if is_master:
                    self.owner.close_loops()



from io import StringIO
from unittest import skip, TestCase

from cosapp.ports import Port, Scope
from cosapp.systems import System

from ..prettyprint import list_inputs, list_outputs


class TestPrettyPrint(TestCase):
    class TestPort(Port):
        def setup(self):
            self.add_variable("Pt", 101325.0, limits=(0.0, None))
            self.add_variable("W", 1.0, valid_range=(0.0, None))

    class SubSystem(System):
        def setup(self):
            self.add_input(TestPrettyPrint.TestPort, "in_")
            self.add_inward(
                "sloss",
                0.95,
                unit="m/s",
                dtype=float,
                valid_range=(0.8, 1.0),
                invalid_comment="not valid",
                limits=(0.0, 1.0),
                out_of_limits_comment="hasta la vista baby",
                desc="get down",
                scope=Scope.PROTECTED,
            )
            self.add_output(TestPrettyPrint.TestPort, "out")
            self.add_outward(
                "tmp",
                unit="inch/lbm",
                dtype=(int, float, complex),
                valid_range=(1, 2),
                invalid_comment="not valid tmp",
                limits=(0, 3),
                out_of_limits_comment="I'll be back",
                desc="banana",
                scope=Scope.PROTECTED,
            )

            self.add_outward("dummy", 1.0)
            self.add_equation("dummy == 0")

        def compute(self):
            for name in self.out:
                self.out[name] = self.in_[name] * self.sloss

            self.dummy /= 100.0

    class TopSystem(System):

        tags = ["cosapp", "tester"]

        def setup(self):
            self.add_inward("top_k")
            self.add_outward("top_tmp")

            self.add_child(
                TestPrettyPrint.SubSystem("sub"), pulling={"in_": "in_", "out": "out"}
            )

    def test_list_residuals(self):
        s = TestPrettyPrint.TopSystem("test")
        outputs = list_outputs(s, residuals=True)
        self.assertEqual(
            sorted(outputs),
            [
                ("test.sub.out.Pt", {"resids": "", "value": 101325.0}),
                ("test.sub.out.W", {"resids": "", "value": 1.0}),
                ("test.sub.residue.dummy == 0", {"resids": 1.0, "value": ""}),
            ],
        )

    def test_list_locals(self):
        s = TestPrettyPrint.TopSystem("test")
        outputs = list_outputs(s, local=True)
        self.assertEqual(
            sorted(outputs),
            [
                ("test.sub.dummy", {"value": 1.0}),
                ("test.sub.out.Pt", {"value": 101325.0}),
                ("test.sub.out.W", {"value": 1.0}),
                ("test.sub.tmp", {"value": 1}),
            ],
        )

    def test_list_hierarchical(self):
        s = TestPrettyPrint.TopSystem("test")
        inputs = list_inputs(s, hierarchical=False)
        self.assertEqual(
            sorted(inputs),
            [
                ("test.sub.in_.Pt", {"value": 101325.0}),
                ("test.sub.in_.W", {"value": 1.0}),
            ],
        )

    def test_list_data(self):
        s = TestPrettyPrint.TopSystem("test")
        inputs = list_inputs(s, inwards=True)
        self.assertEqual(
            sorted(inputs),
            [
                ("test.sub.in_.Pt", {"value": 101325.0}),
                ("test.sub.in_.W", {"value": 1.0}),
                ("test.sub.sloss", {"value": 0.95}),
            ],
        )

    def test_list_return_value(self):
        s = TestPrettyPrint.TopSystem("test")
        inputs = list_inputs(s, out_stream=None)
        self.assertEqual(
            sorted(inputs),
            [
                ("test.sub.in_.Pt", {"value": 101325.0}),
                ("test.sub.in_.W", {"value": 1.0}),
            ],
        )

        # list explicit outputs
        outputs = list_outputs(s, out_stream=None)
        self.assertEqual(
            sorted(outputs),
            [
                ("test.sub.out.Pt", {"value": 101325.0}),
                ("test.sub.out.W", {"value": 1.0}),
            ],
        )

    @skip("TODO")
    def test_list_no_values(self):
        #  check if warning is displayed
        pass

    def test_simple_list_vars_options(self):
        s = TestPrettyPrint.TopSystem("test")
        # list_inputs test
        stream = StringIO()
        inputs = list_inputs(s, out_stream=stream)
        text = stream.getvalue()
        self.assertEqual(
            sorted(inputs),
            [
                ("test.sub.in_.Pt", {"value": 101325.0}),
                ("test.sub.in_.W", {"value": 1.0}),
            ],
        )
        self.assertEqual(1, text.count("2 Input(s) in 'model'"))
        self.assertEqual(1, text.count("top"))
        self.assertEqual(1, text.count("  test"))
        self.assertEqual(1, text.count("    sub"))
        self.assertEqual(1, text.count("      in_"))
        self.assertEqual(1, text.count("        Pt"))
        num_non_empty_lines = sum([1 for s in text.splitlines() if s.strip()])
        self.assertEqual(num_non_empty_lines, 10)

        # list_outputs test
        stream = StringIO()
        outputs = list_outputs(s, out_stream=stream)
        text = stream.getvalue()
        self.assertEqual(
            sorted(outputs),
            [
                ("test.sub.out.Pt", {"value": 101325.0}),
                ("test.sub.out.W", {"value": 1.0}),
            ],
        )
        self.assertEqual(1, text.count("2 Output(s) in 'model'"))
        self.assertEqual(1, text.count("top"))
        self.assertEqual(1, text.count("  test"))
        self.assertEqual(1, text.count("    sub"))
        self.assertEqual(1, text.count("      out"))
        self.assertEqual(1, text.count("        Pt"))
        num_non_empty_lines = sum([1 for s in text.splitlines() if s.strip()])
        self.assertEqual(num_non_empty_lines, 10)




"""
The base code for this Signal/Slot implementation comes from
source: https://github.com/Numergy/signalslot
Copyright (c) Numergy
"""

from cosapp.core.signal.signal import Signal
from cosapp.core.signal.slot import Slot

__all__ = ["Signal", "Slot"]



import json
import logging
import socket
import subprocess
import sys
from abc import abstractmethod
from struct import pack, unpack
from threading import Timer
from typing import Any, Dict, Optional

import numpy.distutils
from numpy.distutils.exec_command import find_executable

from cosapp.systems.system import System
from cosapp.utils.json import JSONEncoder

logger = logging.getLogger(__name__)


class Communication:
    def __init__(self, port):
        self.port = port
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.my_connection = None

    def send_message(self, message):
        msg = message.encode()
        length = pack(">Q", len(msg))
        self.my_connection.sendall(length)
        self.my_connection.sendall(msg)

        # self.my_connection.send(message.encode())

    def wait_for_message(self, nb_characters=1024):
        bs = self.my_connection.recv(8)
        (length,) = unpack(">Q", bs)
        message = b""
        while len(message) < length:
            to_read = length - len(message)
            message += self.my_connection.recv(min(to_read, 4096))

        # message = self.my_connection.recv(nb_characters)
        return message.decode()

    def close_connection(self):
        self.connection.close()

    def __del__(self):
        self.close_connection()


class Server(Communication):
    def __init__(self, port, failed_connection_attempt_max=1):
        Communication.__init__(self, port)

        self.connection.bind(("", port))
        self.connection.listen(failed_connection_attempt_max)

    def accept(self):
        self.my_connection, (clientsocket, ip) = self.connection.accept()
        return self.my_connection, (clientsocket, ip)


class Client(Communication):
    def __init__(self, name, port):
        Communication.__init__(self, port)

        self.name = name
        self.my_connection = None
        self.retry = True
        self.connected = False

    def connect_server(self, timeout=5.0):
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connected = False
        self.timer = Timer(timeout, self.stop)
        self.timer.start()

        while self.retry and not self.connected:
            try:
                self.connection.connect(("localhost", self.port))
            except ConnectionRefusedError:
                pass
            else:
                self.connected = True
                self.my_connection = self.connection
                logger.debug(f"Connected to service provided by {self.name!r} at port {self.port}")

        if not self.connected:
            logger.warning(f"Failed to connect service {self.name!r} at port {self.port}")
            raise ConnectionRefusedError(f"Server is not responding after {timeout}s")

    def stop(self) -> None:
        self.retry = False

    def close_connection(self) -> None:
        self.connected = False
        self.retry = True
        self.connection.close()


class ExternalSystem(System):

    __slots__ = ("_process",)

    def __init__(self, name: str, **kwargs):
        self._process = None
        # here _initialize will be called then user setup
        super().__init__(name, **kwargs)

    def serialize_data(self) -> Dict[str, Any]:
        """Serialize all input data into a dictionary"""
        return {name: port.serialize_data() for name, port in self.inputs.items()}

    @abstractmethod
    def send_inputs(self) -> None:
        pass

    @abstractmethod
    def read_outputs(self) -> Any:
        pass


class TCPSystem(ExternalSystem):
    def __init__(
        self,
        name: str,
        init_variables: Optional[dict] = None,
        port: Optional[int] = 13000,
        **kwargs,
    ):
        object.__setattr__(self, "_port", port)
        object.__setattr__(self, "_client", Client(name, port))
        object.__setattr__(
            self, "_service", {"exec": str(), "script": str(), "arguments": list()}
        )

        # here _initialize will be called then user setup
        super().__init__(name, init_variables, **kwargs)

    def call_setup_run(self):
        super().call_setup_run()
        self._launch_service()
        self._client.connect_server()

        self._client.send_message(self._wrap_inputs())
        msg = self._client.wait_for_message()
        logger.debug(
        	f"Service provided by {self.name!r} is running and returned message {msg!r}"
        )

    def compute(self):
        self._client.send_message(self._wrap_inputs())
        out = self._client.wait_for_message()

        if out == "see you soon":
            logger.debug(
            	f"Successfully disconnected from service provided by {self.name!r}"
            )
            self._client.close_connection()

    def call_clean_run(self):
        try:
            self.close_service()
        except:
            pass
        else:
            self._client.close_connection()
        self._process.terminate()
        super().call_clean_run()

    def close_service(self):
        self._client.send_message("shutdown_service")
        out = self._client.wait_for_message()
        logger.debug(
        	f"Successfully disconnected from service provided by {self.name!r}"
        )

    def _wrap_inputs(self) -> str:
        return json.dumps(self.serialize_data(), cls=JSONEncoder)

    def _launch_service(self) -> None:
        # Make sure command exists
        service = self._service
        command = service['exec']
        if not isinstance(command, str):
            raise TypeError(
                f"Executable of external system {self.name!r} must be a string; got {command!r}")

        if not command:
            raise TypeError(f"Executable of external system {self.name!r} is not defined")

        # Suppress message from find_executable function, we'll handle it
        numpy.distutils.log.set_verbosity(-1)

        command_full_path = find_executable(command)
        if not command_full_path:
            raise ValueError(f"Requested command '{command}' cannot be found")

        command_for_shell_proc = [command, service['script']] + service['arguments']

        if sys.platform == "win32":
            command_for_shell_proc = ["cmd.exe", "/c"] + command_for_shell_proc

        self._process = subprocess.Popen(command_for_shell_proc,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def send_inputs(self) -> None:
        # TODO use this method to send inputs
        pass

    def read_outputs(self) -> Any:
        # TODO use this method to read outputs
        pass



import pytest

import numpy as np

from cosapp.core.eval_str import AssignString
from cosapp.ports.enum import PortType


@pytest.mark.parametrize("lhs", [
    "x[2]",
    "x[-1]",
    "inwards.x[-1]",
    ])
def test_AssignString__init__(eval_context, lhs):
    context = eval_context  # for convenience
    assert context is eval_context
    context.x[2] = 0.1
    assert context.x[2] == 0.1

    rhs = 'a + b'
    s = AssignString(lhs, rhs, context)
    assert s.lhs == lhs
    assert s.rhs == rhs
    assert str(s) == f"{lhs} = {rhs}"
    assert s.eval_context is context
    assert context.x[2] == 0.1
    
    context.a = 0.9
    context.b = 1.7
    value, changed = s.exec()
    assert value == pytest.approx(2.6, rel=1e-14)
    assert context.x[2] == value
    assert changed
    
    context.a = 0.1
    context.b = 0.3
    value, changed = s.exec()
    assert changed
    assert value == pytest.approx(0.4, rel=1e-14)
    assert context.x[2] == value


@pytest.mark.parametrize("lhs, rhs, exception", [
    (0, 0, ValueError),
    ("0", 0, ValueError),
    ("0", "0", ValueError),
    ("0", "a + b", ValueError),
    ("a + b", "0", SyntaxError),
    ("-", "0", SyntaxError),
    ("x[", "0", SyntaxError),
    ("a", "b +", SyntaxError),
    ("a[:]", "b", TypeError),
    ("_a", "b", NameError),
    ("a", "-", SyntaxError),
    ("a + b", "0", SyntaxError),
    ("a + x", "0", SyntaxError),
    ("x", "MyFunc(b)", NameError),  # note: when lhs is a vector, rhs is evaluated at construction
    ("x", "_b", NameError),
    ("x[1:]", "_b", NameError),
    ("x", "cos()", ValueError),
    ("x", "cos(", SyntaxError),
    ("x[1:]", "ones(3)", ValueError),
    ("x", "ones(12)", ValueError),
    ("x", "ones(2)", ValueError),
    ("x[1:]", "[a, _b]", NameError),
    ("a", "_b", NameError),
    ("a", "MyFunc(b)", NameError),
    ("a", "_", NameError),
    ("x[1]", "MyFunc(b)", NameError),
    ])
def test_AssignString__init__error(eval_context, lhs, rhs, exception):
    """Test expressions expected to raise an exception at instantiation"""
    with pytest.raises(exception):
        AssignString(lhs, rhs, eval_context)

@pytest.mark.parametrize("lhs, rhs, exception", [
    ("a", [1], TypeError),
    ("a", "[1]", TypeError),
    ("a", "(1, 2)", TypeError),
    ])
def test_AssignString_exec_error(eval_context, lhs, rhs, exception):
    """Test expressions expected to raise an exception at execution"""
    s = AssignString(lhs, rhs, eval_context)
    with pytest.raises(exception):
        s.exec()


@pytest.mark.parametrize("lhs", ["inwards.a", "a"])
def test_AssignString_exec_a(eval_context, lhs):
    context = eval_context  # for convenience
    assert context is eval_context
    context.a = 0.123
    context.x = np.r_[0.1, 0.2, 0.3]

    rhs = 'b * x[0]'
    s = AssignString(lhs, rhs, context)
    assert s.lhs == lhs
    assert s.rhs == rhs
    assert str(s) == f"{lhs} = {rhs}"
    assert s.eval_context is context
    assert s.shape is None
    assert context.a == 0.123, "The creation of an AssignString should not cause any assignment"
    context.b = 3.14

    # first execution
    value, changed = s.exec()
    assert changed
    assert value == pytest.approx(0.314)
    with pytest.raises(NameError, match="'a' is not defined"):
        assert a == pytest.approx(value)
    assert context.a == pytest.approx(value)

    # second execution
    value, changed = s.exec()
    assert not changed
    assert value == pytest.approx(0.314)


@pytest.mark.parametrize("lhs", ["inwards.x", "x", "x[:]", "x[::]"])
@pytest.mark.parametrize("rhs, setup, expected", [
    ("array([a, b, -a])", {'a': 0.99, 'b': 3.14}, [0.99, 3.14, -0.99]),
    ("[a, b, -a]", {'a': 0.99, 'b': 3.14}, [0.99, 3.14, -0.99]),
    ("a", {'a': 0.99, 'b': 3.14}, np.full(3, 0.99)),
    ("8", {'a': 0.99, 'b': 3.14}, np.full(3, 8.0)),
    (8, {'a': 0.99, 'b': 3.14}, np.full(3, 8.0)),
    ([0.3, 0.2, 0.1], {}, [0.3, 0.2, 0.1]),
    ("[0.3, 0.2, 0.1]", {}, [0.3, 0.2, 0.1]),
    (np.r_[0.3, 0.2, 0.1], {}, [0.3, 0.2, 0.1]),
    # integer-valued rhs: should not change x.dtype
    ([0, 2, 1], {}, [0, 2, 1]),
    (np.ones(3, dtype=int), {}, [1, 1, 1]),
    ])
def test_AssignString_exec_full_array(eval_context, setup, lhs, rhs, expected):
    context = eval_context  # for convenience
    assert context is eval_context
    context.x = np.r_[0.1, 0.2, 0.3]
    assert context.x == pytest.approx([0.1, 0.2, 0.3])
    assert context.x.dtype is np.dtype(float)

    s = AssignString(lhs, rhs, context)
    assert s.shape == (3,)
    assert s.eval_context is context
    assert context.x == pytest.approx([0.1, 0.2, 0.3])

    # setup sytem values
    for name, value in setup.items():
        context[name] = value

    # first execution
    value, changed = s.exec()
    assert changed
    assert value == pytest.approx(expected)
    assert context.x == pytest.approx(expected)
    assert context.x.dtype is np.dtype(float)
    with pytest.raises(NameError, match="'x' is not defined"):
        assert x == pytest.approx(expected)

    # second execution
    value, changed = s.exec()
    assert not changed
    assert value == pytest.approx(expected)
    assert context.x == pytest.approx(expected)


def test_AssignString_array_copy(eval_context):
    """
    Check that an AssignString of the kind `array1 = array2`
    does not make array1 a reference to array2, but assigns a copy instead.
    """
    context = eval_context  # for convenience
    assert context is eval_context
    context.x = np.r_[0.1, 0.2, 0.3]
    context.y = np.zeros(3)

    s = AssignString('x', 'y', context)
    assert s.shape == (3,)
    assert s.eval_context is context
    assert context.x is not context.y

    value, changed = s.exec()
    assert changed
    assert context.x is not context.y
    assert np.array_equal(value, context.x)
    assert np.array_equal(value, context.y)


@pytest.mark.parametrize("lhs, rhs, setup, expected", [
    ("x[::2]", "[a, b]", {'a': 0.99, 'b': 3.14}, [0.99, 0, 3.14]),
    ("x[[0, 2]]", "[a, b]", {'a': 0.99, 'b': 3.14}, [0.99, 0, 3.14]),
    ("x[1:]", "[a, b]", {'a': 0.99, 'b': 3.14}, [0, 0.99, 3.14]),
    ("x[[2, 0, 1]]", "[a, a + 1, a + 2]", {'a': 0.5}, [1.5, 2.5, 0.5]),
    ("x[1:2]", 8, {'a': 0.99, 'b': 3.14}, [0, 8, 0]),
    ("x[1:2]", "8", {'a': 0.99, 'b': 3.14}, [0, 8, 0]),
    ("x[:-1]", "a + b", {'a': 0.99, 'b': 3.14}, [4.13, 4.13, 0]),
    ("x[::2]", np.r_[-2.6, 0.66], {}, [-2.6, 0, 0.66]),
])
def test_AssignString_exec_masked_array(eval_context, setup, lhs, rhs, expected):
    context = eval_context  # for convenience
    assert context is eval_context
    context.x = np.zeros(3)
    assert context.x == pytest.approx([0, 0, 0])

    s = AssignString(lhs, rhs, context)
    assert s.shape is not None
    assert s.eval_context is context
    assert context.x == pytest.approx([0, 0, 0])

    # setup sytem values
    for name, value in setup.items():
        context[name] = value

    # first execution
    value, changed = s.exec()
    assert changed
    assert context.x == pytest.approx(expected)
    with pytest.raises(NameError, match="'x' is not defined"):
        assert x == pytest.approx(expected)

    # second execution
    value, changed = s.exec()
    assert not changed
    assert context.x == pytest.approx(expected)


def test_AssignString_rhs(eval_context):
    context = eval_context
    s = AssignString("sub.in_.q", 0, context)
    assert s.rhs == "0"
    context.sub.in_.q = 2.3
    value, changed = s.exec()
    assert changed
    assert context.sub.in_.q == 0

    s.rhs = "-cos(a)"
    context.a = 1.0
    assert context.sub.in_.q == 0
    value, changed = s.exec()
    assert changed
    assert context.sub.in_.q == -np.cos(1.0)
    context.a = 0.0
    value, changed = s.exec()
    assert changed
    assert context.sub.in_.q == pytest.approx(-1.0)

    with pytest.raises(NameError, match="'foo' is not defined"):
        s.rhs = "2 * foo.bar"

    with pytest.raises(SyntaxError):
        s.rhs = "2 * sin("


@pytest.mark.parametrize("lhs, rhs, expected", [
    ("a", "b", False),
    ("a", "0", True),
    ("a", "cos(pi)", True),
    ("x[1]", "1 + a", False),
    ("x[1]", "1 + a - a", False),
    ("x", "[1, 2, 3]", True),
    ("x", "[1, 2, a]", False),
])
def test_AssignString_constant(eval_context, lhs, rhs, expected):
    """Test whether rhs of assignment is a constant expression"""
    s = AssignString(lhs, rhs, eval_context)
    assert s.constant == expected


def test_AssignString_exec_changed_full_array(eval_context):
    context = eval_context  # for convenience
    assert context is eval_context
    context.x = np.zeros(3)
    assert context.x == pytest.approx([0, 0, 0])

    s = AssignString("x", "[a, a + b, b]", context)
    assert s.shape == (3,)
    assert s.eval_context is context
    assert context.x == pytest.approx([0, 0, 0])

    # setup sytem values
    setup = {'a': 0.99, 'b': 3.14}
    for name, value in setup.items():
        context[name] = value

    # first execution
    value, changed = s.exec()
    assert changed
    assert context.x == pytest.approx([0.99, 4.13, 3.14])
    assert value == pytest.approx(context.x)

    # second execution
    value, changed = s.exec()
    assert not changed
    assert context.x == pytest.approx([0.99, 4.13, 3.14])
    assert value == pytest.approx(context.x)

    # third execution
    context.x[1] = 0  # change x outside of AssignString s
    assert context.x == pytest.approx([0.99, 0, 3.14])
    value, changed = s.exec()
    assert changed
    assert context.x == pytest.approx([0.99, 4.13, 3.14])
    assert value == pytest.approx(context.x)


def test_AssignString_exec_changed_masked_array(eval_context):
    context = eval_context  # for convenience
    assert context is eval_context
    context.x = np.zeros(3)
    assert context.x == pytest.approx([0, 0, 0])

    s = AssignString("x[::2]", "[a, b]", context)
    assert s.shape is not None
    assert s.eval_context is context
    assert context.x == pytest.approx([0, 0, 0])

    # setup sytem values
    setup = {'a': 0.99, 'b': 3.14}
    for name, value in setup.items():
        context[name] = value

    # first execution
    value, changed = s.exec()
    assert changed
    assert context.x == pytest.approx([0.99, 0, 3.14])

    # second execution
    context.x[1] = 2.4
    value, changed = s.exec()
    assert not changed  # Note: x[1] has changed, but s detects no changes on x[::2]
    assert context.x == pytest.approx([0.99, 2.4, 3.14])

    # Change a and the expression of s.rhs, such that the actual value is unchanged
    context.a = 0.33
    s.rhs = "[3 * a, b]"
    value, changed = s.exec()
    assert not changed
    assert context.x == pytest.approx([0.99, 2.4, 3.14])

    context.a = 0.15
    value, changed = s.exec()
    assert changed
    assert value == pytest.approx([0.45, 3.14])
    assert context.x == pytest.approx([0.45, 2.4, 3.14])

    context.set_clean(PortType.IN)
    value, changed = s.exec()
    assert not changed
    assert value == pytest.approx([0.45, 3.14])
    assert context.x == pytest.approx([0.45, 2.4, 3.14])
    assert context.is_clean(PortType.IN)  ### Assignment does *not* modify clean/dirty status

    context.set_clean(PortType.IN)
    context.x = np.zeros(3)  # change x outside of AssignString s
    assert not context.is_clean(PortType.IN)
    value, changed = s.exec()
    assert changed
    assert value == pytest.approx([0.45, 3.14])
    assert context.x == pytest.approx([0.45, 0, 3.14])

    context.x[::2] = np.ones(2)  # change x outside of AssignString s
    assert context.x == pytest.approx([1, 0, 1])
    value, changed = s.exec()
    assert changed
    assert value == pytest.approx([0.45, 3.14])
    assert context.x == pytest.approx([0.45, 0, 3.14])


@pytest.mark.parametrize("rhs, value", [
    ("0", 0),
    ("1 + 4", 5),
    ("1.23 / 10", 1.23 / 10),
    ("cos(pi)", np.cos(np.pi)),
    ("26 * pi / 180", 26 * np.pi / 180),
    ("exp(-1.5)", np.exp(-1.5)),
    ("log(2)", np.log(2)),
    ])
def test_AssignString_constant_evaluation(eval_context, rhs, value):
    """Test that constant expressions are evaluated before being stored"""
    s = AssignString("a", rhs, eval_context)
    assert s.constant
    assert s.rhs == rhs  # representation of rhs is unchanged
    assert str(s._AssignString__sides) == f"(a, {value})"  # actual value is stored in 'sides'



import pytest

from cosapp.drivers import RunOnce, NonLinearSolver
from cosapp.tests.library.systems import ComplexDuct


def test_ComplexDuct():
    s = ComplexDuct("MyComplexDuct")

    s.fl_in.W = 100
    s.fl_in.Pt = 1020.408163265
    s.fl_in.Tt = 50.
    s.run_once()

    s.bleed.fl2_out.Pt = 1000

    s.add_driver(RunOnce("first run"))
    s.add_driver(NonLinearSolver("design"))
    s.run_drivers()
    assert s.bleed.fl2_out.W == pytest.approx(100 / 99, rel=1e-5)
    assert s.merger.fl2_in.W == pytest.approx(100 / 99, rel=1e-5)
    assert s.bleed.fl2_out.Pt == pytest.approx(1000, rel=1e-5)
    assert s.merger.fl2_in.Pt == pytest.approx(1000, rel=1e-5)
